# coding: utf-8
#
# Copyright (c) 2024 National Technology and Engineering Solutions of
# Sandia, LLC. Under the terms of Contract DE-NA0003525 with NTESS,
# the U.S. Government retains certain rights in this software.
#
# SPDX-License-Identifier: BSD-3-Clause.

"""
The SANSMIC main program.

This submodule provides the hook for the command line programs ``sansmic``
and ``sansmic-convert``.
"""

import logging
from argparse import ArgumentParser, Action
from os.path import splitext
import sys

from numpy import round
from pip._vendor.rich.progress import Progress
import sansmic.io

try:
    import h5py
except ImportError as e:
    h5py = e

logging.basicConfig()
logger = logging.getLogger("sansmic")


class LicenseAction(Action):
    def __call__(default, parser, namespace, values, option_string=None):
        if values != "bar":
            from sansmic import __license__

            print(__license__)
            parser.exit(0)


class CopyrightAction(Action):
    def __call__(default, parser, namespace, values, option_string=None):
        if values != "bar":
            from sansmic import __copyright__

            print(__copyright__)
            parser.exit(0)


def _get_verbosity(args):
    """Set the logger level for sansmic."""
    level = logging.WARNING
    if args.verbose:
        level = logging.INFO
    elif args.quiet:
        level = logging.ERROR
    if args.debug:
        level = logging.DEBUG
    logger.setLevel(level)
    if args.debug:
        args.verbose = 3
        return 3
    return args.verbose


def _main_parser(defaults=False):
    from sansmic import __version__

    if not defaults:
        kwargs = dict(default=None)
    else:
        kwargs = dict()
    parser = ArgumentParser(
        prog=f"sansmic v.{__version__}",
        description="Simulate leaching in an underground salt cavern due to ordinary solution mining, product withdrawal, or product fill.",
        epilog=f"sansmic (c) 2024 NTESS. Use the --copyright or --license flags for full details.",
    )
    parser.add_argument(
        "--copyright",
        default=None,
        nargs=0,
        action=CopyrightAction,
        help="Display full copyright details and exit.",
    )
    parser.add_argument(
        "--license",
        default=None,
        nargs=0,
        action=LicenseAction,
        help="Display full license details and exit.",
    )
    parser.add_argument(
        "datafile",
        metavar="INPUTFILE",
        help="the SANSMIC input file to use; if extensansmic.ion is '.dat', --toml is implied",
    )

    outp = parser.add_argument_group(
        "Output options",
        """Default options:  ``--csv  --tst  --no-hdf  --no-json  --no-old-out``""",
    )
    outp.add_argument(
        "-o",
        "--output",
        dest="prefix",
        help="change output file prefix; if used, --toml is implied",
        default=None,
    )

    wwo = outp.add_mutually_exclusive_group()
    wwo.add_argument(
        "--toml",
        default=None,
        action="store_true",
        help="Create a TOML scenario file.",
    )
    wwo.add_argument(
        "--no-toml",
        dest="toml",
        action="store_false",
        help="Don't create a TOML scenario file.",
        **kwargs,
    )

    wwo = outp.add_mutually_exclusive_group()
    wwo.add_argument(
        "--csv",
        action="store_true",
        default=True if defaults else None,
        help="Create CSV results files.",
    )
    wwo.add_argument(
        "--no-csv",
        dest="csv",
        action="store_false",
        help="Don't create CSV results files.",
        **kwargs,
    )

    wwo = outp.add_mutually_exclusive_group()
    wwo.add_argument(
        "--hdf",
        default=False if defaults else None,
        action="store_true",
        help="Create an HDF5 formatted results file.",
    )
    wwo.add_argument(
        "--no-hdf",
        dest="hdf",
        action="store_false",
        help="Don't create an HDF5 results file.",
        **kwargs,
    )

    wwo = outp.add_mutually_exclusive_group()
    wwo.add_argument(
        "--json",
        default=False if defaults else None,
        action="store_true",
        help="Create a JSON formatted results file.",
    )
    wwo.add_argument(
        "--no-json",
        dest="json",
        action="store_false",
        help="Don't create a JSON results file.",
        **kwargs,
    )

    wwo = outp.add_mutually_exclusive_group()
    wwo.add_argument(
        "--tst",
        default=True if defaults else None,
        action="store_true",
        help="Create a daily summary TST text file.",
    )
    wwo.add_argument(
        "--no-tst",
        dest="tst",
        action="store_false",
        help="Don't create a TST summary file.",
        **kwargs,
    )

    wwo = outp.add_mutually_exclusive_group()
    wwo.add_argument(
        "--old-out",
        dest="old_out",
        action="store_true",
        default=False if defaults else None,
        help="Create an old-style SANSMIC OUT file.",
    )
    wwo.add_argument(
        "--no-old-out",
        dest="old_out",
        action="store_false",
        help="Don't create an old-style OUT file.",
        **kwargs,
    )

    outp = parser.add_argument_group(
        "Console/stdout/stderr reporting",
        "The following options are mutually exclusive",
    )
    verb = outp.add_mutually_exclusive_group()
    verb.add_argument(
        "-v",
        "--verbose",
        help="increase reporting details",
        action="count",
        default=0,
    )
    verb.add_argument(
        "-q",
        "--quiet",
        help="turn off runtime screen output",
        action="store_true",
        default=False,
    )
    verb.add_argument(
        "--debug",
        help="turn on debug messages and set max verbosity",
        action="store_true",
        default=False,
    )
    return parser


def main(args=None, ret=False):
    """Command line function to run sansmic.

    If this function is run from somewhere other than the command line, then
    arguments can be passed in to control execution as a list of strings.
    The ``ret`` argument can be set to ``True`` to return the results when
    using this as a function, also.

    Parameters
    ----------
    args : argparse.Namespace or list[str]
        Arguments as a list of strings that would be used on the command-line.
    ret : bool
        Should the function return a results object, by default False.

    """
    extra_args = args is not None
    if ret or args is not None:
        defaults = False
    else:
        defaults = True
    parser = _main_parser(defaults)
    args = parser.parse_args(args=args)

    # Wrap the different sections in try/except blocks
    try:
        datafile = args.datafile
        prefix = splitext(datafile)[0] if args.prefix is None else args.prefix
        verbosity = _get_verbosity(args)
        logger.debug("Running sansmic with {}".format(args))
        model = sansmic.io.read_scenario(datafile, warn=not args.quiet)
        logger.info("Successfully created scenario from {}".format(datafile))
        if not args.toml:
            pass
        elif (
            args.toml is None and args.prefix is not None
        ) or datafile.lower().endswith(".dat"):
            args.toml = True
        elif args.toml is None and args.prefix is None:
            args.toml = False
        if args.toml:
            sansmic.io.write_scenario(model, prefix + ".toml")
    except Exception as e:
        if extra_args:
            raise e
        parser.error(str(e))

    logger.debug("Running simulation")
    with model.new_simulation(prefix, verbosity, args.tst, args.old_out) as sim:
        logger.debug("Created new simulation")
        if args.quiet or not args.verbose:
            logger.debug("Running in batch mode")
            if not args.quiet:
                print("Running sansmic in batch mode from ".format(repr(datafile)))
            sim.run_sim()
        else:
            logger.debug("Running in stepwise mode")
            last_stage = 0
            last_step = 0
            n_stages = len(model.stages)
            dt_stage = [s.solver_timestep for s in model.stages]
            day_size = [int(round(24.0 / dt)) for dt in dt_stage]
            t_stage = [s.injection_duration + s.rest_duration for s in model.stages]
            t_inject = [s.injection_duration for s in model.stages]
            stage_sizes = [
                int(round(t_stage[ct] / dt_stage[ct])) for ct in range(n_stages)
            ]
            n_steps = sum(stage_sizes)
            p_freq = day_size[0]
            print(
                "Running sansmic scenario: {}".format(
                    datafile if not model.title else model.title
                )
            )
            stage = 0
            with Progress() as progress:
                if args.verbose >= 1:
                    task = progress.add_task("Progress...", total=n_steps)
                if args.verbose >= 2:
                    task_S = progress.add_task(
                        "[red]Stage {}...".format(stage + 1), total=stage_sizes[stage]
                    )
                for stage, step in sim.steps:
                    if last_stage != stage:
                        if stage >= len(model.stages):
                            if args.verbose >= 1:
                                progress.update(
                                    task,
                                    completed=n_steps,
                                )
                            if args.verbose >= 2:
                                progress.update(
                                    task_S,
                                    completed=n_steps,
                                )
                                print("All stages complete.")
                        else:
                            last_stage = stage
                            last_step = step
                            p_freq = day_size[stage]
                            if args.verbose >= 2:
                                progress.update(
                                    task_S,
                                    completed=n_steps,
                                )
                                task_S = progress.add_task(
                                    "[red]Stage {}...".format(stage + 1),
                                    total=stage_sizes[stage],
                                )
                    else:
                        if args.verbose >= 1 and (step - last_step) % p_freq == 0:
                            progress.update(
                                task,
                                advance=p_freq,
                            )
                        if args.verbose >= 2 and (step - last_step) % p_freq == 0:
                            progress.update(
                                task_S,
                                advance=p_freq,
                            )
        logger.debug("Simulation complete")
    res = sim.results
    if not args.quiet:
        print("Initial and final results:")
        print(
            (res.df_t_1D.iloc[[0, -1], [1, 3, 13, 15, 19, 20, 21, 26]]).to_string(
                index=False
            )
        )

    # Wrap the outputs in a try/except block
    try:
        if args.csv:
            sansmic.io.write_csv_results(res, prefix)
        if args.json:
            sansmic.io.write_json_results(res, prefix + ".json")
        if args.hdf:
            sansmic.io.write_hdf_results(res, prefix + ".h5")
        logger.debug("Sansmic complete")
    except Exception as e:
        if extra_args:
            raise e
        logger.critical("Error while writing results - some results may be missing")
        parser.error(str(e))

    if ret:
        return res


def _convert_parser():
    from sansmic import __version__

    parser = ArgumentParser(
        prog=f"sansmic-convert v.{__version__}",
        description="Convert from an old-style DAT file to the new TOML format.",
        epilog=f"sansmic (c) 2024 NTESS. Use the --copyright or --license flags for full details.",
    )
    parser.add_argument(
        "--copyright",
        default=None,
        nargs=0,
        action=CopyrightAction,
        help="Display full copyright details and exit.",
    )
    parser.add_argument(
        "--license",
        default=None,
        nargs=0,
        action=LicenseAction,
        help="Display full license details and exit.",
    )
    parser.add_argument(
        "infile", metavar="OLD_FILE", help="the SANSMIC input file to convert"
    )
    parser.add_argument(
        "outfile",
        metavar="NEW_FILE",
        help="the new scenario file to create [extensansmic.ion choices: .toml, .json, .yaml]",
        default=None,
    )
    parser.add_argument(
        "-f",
        "--full",
        action="store_true",
        default=False,
        help="Keep all entries, even if they are blank",
    )
    return parser


def convert(args=None):
    """Command line function to convert a scenario/dat to a new format."""
    extra_args = args is not None
    parser = _convert_parser()
    args = parser.parse_args(args=args)
    if args.license:
        from sansmic import __license__

        print(__license__)
        parser.exit(0)
    elif args.copyright:
        from sansmic import __copyright__

        print(__copyright__)
        parser.exit(0)
    infile = args.infile
    logger.debug("Running sansmic-convert")
    try:
        model = sansmic.io.read_scenario(infile)
        logger.debug("Successfully created scenario from {}".format(infile))
        sansmic.io.write_scenario(model, args.outfile, redundant=args.full)
    except Exception as e:
        if extra_args:
            raise e
        parser.error(str(e))
    logger.debug("Successfully wrote scenario to {}".format(args.outfile))



# coding: utf-8
#
# Copyright (c) 2024 National Technology and Engineering Solutions of
# Sandia, LLC. Under the terms of Contract DE-NA0003525 with NTESS,
# the U.S. Government retains certain rights in this software.
#
# SPDX-License-Identifier: BSD-3-Clause.

"""The core python classes for SANSMIC."""

import logging
import warnings
from dataclasses import InitVar, asdict, dataclass, field
from enum import IntEnum
from fractions import Fraction
from typing import Any, Dict, List, Literal, Union

import numpy as np
import pandas as pd

try:
    import h5py
except ImportError as e:
    h5py = e

from . import libsansmic as _ext

logger = logging.getLogger("sansmic")


def _rename_with_underscore(orig: dict, new: dict):
    """Rename all keys in a dictionary with underscores instead of spaces or dashes"""
    for k, v in orig.items():
        k = k.lower().replace("-", "_")
        new[k] = v


def _rename_with_dash(orig: dict, new: dict):
    """Rename all keys in a dictionary with dashes instead of underscores"""
    for k, v in orig.items():
        k = k.lower().replace("_", "-")
        new[k] = v


class Units(IntEnum):
    """The units that are used to define the scenario.

    Depths, heights, and cavern radii are in the first (larger) length unit.
    Tubing radii are in the second (smaller) length unit.
    Volumes and volumetric flow rates are in the third unit.

    Durations are in hours. Constant injection rates are in <volume-unit> per day.
    File-based injection rates are in <volume-unit> per hour.
    """

    FT_IN_BBL = 1
    """foot/inch/barrel"""
    FT_IN_FT3 = 2
    """foot/inch/cubic foot"""
    M_CM_M3 = 3
    """meter/centimeter/cubic meter"""

    @classmethod
    def inch(self):
        """1 in ≔ 0.0254 m"""
        return Fraction(254, 10000)

    @classmethod
    def foot(self):
        """1 ft ≔ 0.3048 m"""
        return Fraction(3048, 10000)

    @classmethod
    def cubic_foot(self):
        """1 ft³ == 0.028316846592 m³"""
        return Fraction(3048**3, 10000**3)

    @classmethod
    def barrel(self):
        """1 bbl == 0.158987294928 m³"""
        return Fraction(42 * 231 * 254, 10000**3)

    @classmethod
    def centimeter(self):
        """1 cm ≔ 0.01 m"""
        return Fraction(1, 100)


class StopCondition(IntEnum):
    """Which stop condition should be used to end the simulation."""

    DEPTH = -1
    """Stop when the interface reaches a specified height above the cavern floor"""
    DURATION = 0
    """Stop only based on duration"""
    VOLUME = 1
    """Stop when the total cavern volume reaches a specified value."""


class GeometryFormat(IntEnum):
    """The format for the initial cavern geometry."""

    RADIUS_LIST = 0
    """The radius is provided from the bottom of the cavern to the top;
    :attr:`~Scenario.num_cells` + 1 values, equally spaced"""
    VOLUME_LIST = 1
    VOLUME_TABLE = -1
    RADIUS_TABLE = 2
    LAYER_CAKE = 5
    """The geometry is provided in a 'layer-cake' style LAS file"""


class SimulationMode(IntEnum):
    """The simulation mode determines which options are active for injection."""

    ORDINARY = 0
    """Ordinary leaching, with raw water or undersaturated brine injected through
    the inner tubing or outer casing and brine produced from the other."""
    WITHDRAWAL = 1
    """Withdrawal leach, with brine injected through the suspended tubing (hanging
    string) and product produced from top of cavern."""
    LEACH_FILL = 2
    """Simultaneous leaching (water/brine injection and brine production) and
    product injection from the top."""
    STORAGE_FILL = -1
    """Storage fill, with product injected from the top of the cavern and brine
    produced from the suspended tubing/hanging string."""


class RateScheduleType(IntEnum):
    """The way that the rate of injection is specified."""

    CONSTANT_RATE = 1
    """Injection occurs at a constant rate of volume :class:`Units` per day."""
    DATAFILE = 2
    """Injection rate is specified in :class:`Units` per hour in a file."""


@dataclass
class AdvancedOptions:
    """Advanced configuration options. Most of these can and should be left as None which
    will use the default C++ library values."""

    absolute_error: float = None  #  1.0e-2
    """ODE solver absolute error tolerance; CScenario default is 0.01"""
    relative_error: float = None  #  1.0e-4
    """ODE solver absolute error tolerance; CScenario default is 0.0001"""
    coallescing_wells: int = None  #  1
    """The number of wells for multi-well cavern construction; CScenario default is 1"""
    well_separation: float = None  # 0.0
    """The separation distance for multi-well cavern construction; CScenario default is 0.0"""
    jet_model_version: int = None  #  1
    """Jet model version; CScenario default is 1"""
    plume_model_version: int = None  #  1
    """Plume model version; CScenario default is 1"""
    temperature_model_version: int = None  #  0
    """Temperature model version; CScenario default is 0"""
    dissolution_factor: float = None  # 1.0
    """Dissolution factor; CScenario default is 1.0 - this should not be changed unless you are sure you know the effects"""
    max_brine_sg: float = None  #  1.2019
    """Maximum brine specific gravity; CSalt default is 1.2019"""
    solid_density: float = None  #  2.16
    """Rock density in solid form; CSalt default is 2.16 g/cc"""
    entrainment_coeff: float = None  #  0.09
    """Dissolution entrainment coefficient; CScenario default is 0.09"""
    molecular_diffusion: float = None  #  5.03e-5
    """Molecular diffusion coefficient; CScenario default is 5.03e-5"""
    eddy_coefficient: float = None  #  1.142e5
    """Eddy coefficient; CScenario default is 1.142e5"""
    diffusion_beta: float = None  #  0.147
    """Diffusion beta coefficient; default is 0.147"""

    def __setattr__(self, name, value):
        if isinstance(value, str) and value.strip() == "":
            value = None
        super().__setattr__(name, value)

    @classmethod
    def from_dict(cls, opts: dict) -> "AdvancedOptions":
        """Create a new object from a dictionary of options.

        This method differs from the __init__ constructor by automatically
        converting non-underscore characters - e.g., ``-`` or ``.`` or `` ``
        to underscores and changing keys to lower-case prior to creating
        the object.

        See also: :meth:`~StageDefinition.to_dict`

        Parameters
        ----------
        opts : dict
            The initialization values

        Returns
        -------
        StageDefinition
            the new stage
        """

        new_opts = dict()
        _rename_with_underscore(opts, new_opts)
        return cls(**new_opts)

    def to_dict(self, keep_empty: bool = False):
        """Convert the object's data to a dictionary of options.

        This method differs from the :meth:`~dataclasses.asdict`
        method by automatically converting underscore characters to
        hyphens for a more readable dictionary. Specifically used
        when creating TOML, JSON and YAML files.

        See also: :meth:`~StageDefinition.from_dict`

        Parameters
        ----------
        keep_empty : bool
            Keep values set to None and set them to ``""``

        Returns
        -------
        dict
            the options dictionary
        """

        ret = dict()
        _rename_with_dash(asdict(self), ret)
        keys = list(ret.keys())
        for k in keys:
            if ret[k] is None:
                if keep_empty:
                    ret[k] = ""
                else:
                    del ret[k]
        return ret


@dataclass
class StageDefinition:
    """Defines a SANSMIC simulation stage which is an injection period followed by a rest period."""

    title: str = None
    """Title for the stage."""
    simulation_mode: SimulationMode = None
    """The simulation mode used in this stage."""
    solver_timestep: float = None
    """The solver timestep in hours."""
    save_frequency: Union[int, Literal["hourly", "daily", "bystage"]] = "daily"
    """The save frequency in number of timesteps or one of {"hourly", "daily", "bystage"}, by default "daily"."""
    injection_duration: float = None
    """The duration of the injection phase of the stage."""
    rest_duration: float = None
    """The duration of the post-injection rest phase of the stage."""

    inner_tbg_inside_diam: float = None
    """The inner tubing inner diameter (ID)."""
    inner_tbg_outside_diam: float = None
    """The inner tubing outer diameter (OD)."""
    outer_csg_inside_diam: float = None
    """The outer tubing/casing inner diameter (ID)."""
    outer_csg_outside_diam: float = None
    """The outer tubing/casing outer diameter (OD)."""

    brine_injection_sg: float = None
    """The water/brine injection specific gravity."""
    brine_injection_depth: float = None
    """The depth below surface/:term:`ZDP` where the end of the injection string/casing/tubing is positioned."""
    brine_production_depth: float = None
    """The depth below surface/:term:`ZDP` where the end of the production string/casing/tubing is positioned."""
    brine_injection_rate: Union[float, str] = 0
    """The meaning of this field is based on the type of data that is provided (a value of 0 means off)."""
    # """
    # float : constant rate
    #     Injection occurs at a constant rate for the duration of the
    #     :attr:`injection_duration`.

    # str : file path
    #     Injection rates should be read from a file. The file can be any type that can
    #     be read into a pd DataFrame. An 'index' column is optional, but if not provided,
    #     then all data is assumed to be pre-sorted. In addition to the
    #     'index' column, files may contain exactly one set of the following columns:

    #     ['rate', 'duration']
    #         This style gives a list of rates and how long they should continue for.
    #         Durations are provided in decimal hours. If the sum of durations is greater
    #         than :attr:`injection_duration`, an error will be raised. If the
    #         sum of durations is less than :attr:`injection_duration`, the remaining
    #         time will be assumed to be no-flow.

    #     ['time', 'rate']
    #         This style gives a list of new rate values that start at the given time
    #         after the start of the stage. The 'time' values must be increasing and
    #         never duplicate. Times are in decimal hours since the start of the stage.
    #         The last rate will be used up until the end of the
    #         :attr:`injection_duration`. If a time is given that is after the
    #         :attr:`injection_duration`, then an error will be raised.

    #     ['hourly']
    #         This style specifies a new rate for each hour of the stage. There must be
    #         a value specified for each hour of the :attr`injection_duration`, or an
    #         error will be raised.

    #     The optional column ['sg'] can be provided to change the injection water/brine
    #     specific gravity; however, if provided, it must contain a value for every row.

    #     If 'duration' and 'time' occur in the same file, an error will be raised.
    #     If 'time', 'rate', or 'duration' occur in the same file as 'hourly', then
    #     an error will be raised.

    #     Columns with names other than those described above will be ignored.

    #     .. attention::

    #         In all cases, a no-flow, leaching-only period will occur for
    #         :attr:`rest_duration` hours **after** all injection data
    #         is processed -- this means that if an injection file finishes with 0s
    #         that there will be an additional :attr:`rest_duration` hours added to the
    #         end of the stage.

    #     .. warning::

    #         The same file cnnot be used for both brine injection and product
    #         injection data.
    # """

    set_initial_conditions: bool = None
    """Unlink initial cavern brine gravity and interface level from previous stage.
    Automatically set to True for the first stage added to a model."""
    set_cavern_sg: float = None
    """Set the initial specific gravity for all brine-filled cells of the cavern to
    this value. If set_initial_conditions is False and this is not None (or 0) an error
    will be raised upon scenario validation."""
    brine_interface_depth: float = None
    """Set the initial oil-brine interface or blanket level. By default None, which
    will link to the previous stage (as will a value of 0)."""
    # set_cavern_temp: float = None
    # """Set the initial temperature for all brine-filled cells of the cavern; by
    # default None, which will link to the previous stage."""

    product_injection_rate: Union[float, str] = 0.0
    """Either a constant rate of product injection or a file with an injection schedule, by default 0."""
    # product_water_content: float = None
    # """The volume-percent water within the product; a value of None turns this option off."""
    product_injection_depth: float = None
    """The depth below surface/:term:`ZDP` where the product injection string -- or casing or chimney bottom -- is positioned; a value 0 or None sets this to the roof of the cavern."""
    product_production_depth: float = None
    """The depth below surface/:term:`ZDP` where the product production string -- or casing or chimney bottom -- is positioned; a value 0 or None sets this to the roof of the cavern."""

    stop_condition: StopCondition = StopCondition.DURATION
    """Set an additional stop condition for the stage, see :class:`StopCondition`."""
    stop_value: float = 0
    """If the :attr:`stop_condition` is not :attr:`~StopCondition.DURATION`, then the depth or volume value."""

    defaults: InitVar[Dict[str, Any]] = None

    valid_default_keys = [
        "solver_timestep",
        "save_frequency",
        "inner_tbg_inside_diam",
        "inner_tbg_outside_diam",
        "outer_csg_inside_diam",
        "outer_csg_outside_diam",
    ]

    def __post_init__(self, defaults=None):
        if defaults is None:
            return
        if not isinstance(defaults, dict):
            raise TypeError("defaults must be a dictionary")
        for k, v in defaults.items():
            k2 = k.strip().replace("-", "_").replace(" ", "_").replace(".", "_")
            if k2 not in self.valid_default_keys:
                logger.warning(
                    "Ignoring non-defaultable or unknown setting {} = {}".format(
                        k, repr(v)
                    )
                )
            elif getattr(self, k2) is None:
                setattr(self, k2, v)

    def __setattr__(self, name, value):
        if isinstance(value, str) and value.strip() == "":
            value = None
        if name == "simulation_mode" and not isinstance(value, SimulationMode):
            if isinstance(value, int):
                value = SimulationMode(value)
            elif isinstance(value, str):
                value = SimulationMode[
                    value.upper().replace(" ", "_").replace("-", "_")
                ]
            elif isinstance(value, _ext.CRunMode):
                value = SimulationMode(int(value))
            else:
                TypeError("simulation_mode cannot be of type {}".format(type(value)))
        elif name == "stop_condition" and not isinstance(value, StopCondition):
            if isinstance(value, int):
                value = StopCondition(value)
            elif isinstance(value, str):
                value = StopCondition[value.upper().replace(" ", "_").replace("-", "_")]
            elif isinstance(value, _ext.CGeometryFormat):
                value = StopCondition(int(value))
            else:
                TypeError("stop_condition cannot be of type {}".format(type(value)))
        elif name == "set_cavern_sg" and value is not None and value < 1.0:
            value = None
        elif name == "set_initial_conditions" and value is False:
            self.set_cavern_sg = None
            self.brine_interface_depth = None
        super().__setattr__(name, value)

    @classmethod
    def from_dict(cls, opts: dict, defaults=None) -> "StageDefinition":
        """Create a new object from a dictionary of options.

        This method differs from the __init__ constructor by automatically
        converting non-underscore characters - e.g., ``-`` or ``.`` or `` ``
        to underscores and changing keys to lower-case prior to creating
        the object.

        See also: :meth:`~StageDefinition.to_dict`

        Parameters
        ----------
        opts : dict
            The initialization values

        Returns
        -------
        StageDefinition
            the new stage
        """

        new_opts = dict()
        _rename_with_underscore(opts, new_opts)
        return cls(**new_opts, defaults=defaults)

    def to_dict(self, keep_empty: bool = False):
        """Convert the object's data to a dictionary of options.

        This method differs from the :meth:`~dataclasses.asdict`
        method by automatically converting underscore characters to
        hyphens for a more readable dictionary. Specifically used
        when creating TOML, JSON and YAML files.

        See also: :meth:`~StageDefinition.from_dict`

        Parameters
        ----------
        keep_empty : bool
            Keep values set to None and set them to ``""``

        Returns
        -------
        dict
            the options dictionary
        """

        ret = dict()
        _rename_with_dash(asdict(self), ret)
        keys = list(ret.keys())
        for k in keys:
            if ret[k] is None:
                if keep_empty:
                    ret[k] = ""
                else:
                    del ret[k]
        return ret

    def validate(self):
        """Validate that all required options have been selected.

        Raises
        ------
        TypeError
            if there are missing or misconfigured options
        ValueError
            if timing options are invalid

        """

        if self.solver_timestep and self.solver_timestep <= 0:
            raise ValueError("Timestep must be greater than 0 hours")
        if self.injection_duration <= 0:
            raise ValueError("Injection duration must be greater than 0 hours")
        if self.rest_duration < 0:
            raise ValueError("Rest duration must be positive")

        # Validate appropriate initial conditions settings
        if self.set_initial_conditions and self.set_cavern_sg is None:
            logger.warn(
                "Setting the initial conditions without setting cavern sg -- cavern sg will be set to fully saturated brine"
            )
            self.set_cavern_sg = 10.0
        elif (
            not self.set_initial_conditions
            and self.set_cavern_sg is not None
            and self.set_cavern_sg >= 1.0
        ):
            # raise TypeError(
            logger.warn(
                "Setting the starting cavern sg requires set_initial_conditions to be set to True -- setting to 0.0"
            )
            self.set_cavern_sg = 0.0
        elif not self.set_initial_conditions and self.brine_interface_depth:
            logger.warn(
                "Make sure you meant to reset the interface level; use 0.0 or None to continue from the last stage."
            )

        # Validate appropriate brine injection and production settings
        if (
            self.simulation_mode
            in [
                SimulationMode.ORDINARY,
                SimulationMode.LEACH_FILL,
                SimulationMode.STORAGE_FILL,
            ]
            and self.brine_production_depth is None
        ):
            raise TypeError(
                "Missing required 'brine_production_depth' for {} simulation mode.".format(
                    self.simulation_mode.name
                )
            )
        if (
            self.simulation_mode
            in [
                SimulationMode.ORDINARY,
                SimulationMode.LEACH_FILL,
                SimulationMode.WITHDRAWAL,
            ]
            and self.brine_injection_depth is None
        ):
            raise TypeError(
                "Missing required 'brine_injection_depth' for {} simulation mode.".format(
                    self.simulation_mode.name
                )
            )
        if (
            self.simulation_mode
            in [
                SimulationMode.ORDINARY,
                SimulationMode.LEACH_FILL,
                SimulationMode.WITHDRAWAL,
            ]
            and self.brine_injection_rate is None
        ):
            raise TypeError(
                "Missing required 'brine_injection_rate' for {} simulation mode.".format(
                    self.simulation_mode.name
                )
            )
        if (
            self.simulation_mode
            in [
                SimulationMode.ORDINARY,
                SimulationMode.LEACH_FILL,
                SimulationMode.WITHDRAWAL,
            ]
            and self.brine_injection_sg is None
        ):
            raise TypeError(
                "Missing required 'brine_injection_sg' for {} simulation mode.".format(
                    self.simulation_mode.name
                )
            )

        # Validate casing size information
        if (
            self.simulation_mode
            in [
                SimulationMode.ORDINARY,
                SimulationMode.LEACH_FILL,
                SimulationMode.WITHDRAWAL,
                SimulationMode.STORAGE_FILL,
            ]
            and self.inner_tbg_inside_diam is None
        ):
            raise TypeError(
                "Missing required 'inner_tbg_inside_diam' for {} simulation mode.".format(
                    self.simulation_mode.name
                )
            )
        if (
            self.simulation_mode
            in [
                SimulationMode.ORDINARY,
                SimulationMode.LEACH_FILL,
                SimulationMode.WITHDRAWAL,
                SimulationMode.STORAGE_FILL,
            ]
            and self.inner_tbg_outside_diam is None
        ):
            raise TypeError(
                "Missing required 'inner_tbg_outside_diam' for {} simulation mode.".format(
                    self.simulation_mode.name
                )
            )
        if (
            self.simulation_mode in [SimulationMode.ORDINARY, SimulationMode.LEACH_FILL]
            and self.outer_csg_inside_diam is None
        ):
            raise TypeError(
                "Missing required 'outer_csg_inside_diam' for {} simulation mode.".format(
                    self.simulation_mode.name
                )
            )
        if (
            self.simulation_mode in [SimulationMode.ORDINARY, SimulationMode.LEACH_FILL]
            and self.outer_csg_outside_diam is None
        ):
            raise TypeError(
                "Missing required 'outer_csg_outside_diam' for {} simulation mode.".format(
                    self.simulation_mode.name
                )
            )

        # Product injection depth and rate validation
        # if (
        #     self.simulation_mode in [SimulationMode.LEACH_FILL, SimulationMode.STORAGE_FILL]
        #     and self.product_injection_depth is None
        # ):
        #     self.product_injection_depth = 0.0
        if (
            self.simulation_mode
            in [SimulationMode.LEACH_FILL, SimulationMode.STORAGE_FILL]
            and self.product_injection_rate is None
        ):
            raise TypeError(
                "Missing required 'product_injection_rate' options for {} simulation mode.".format(
                    self.simulation_mode.name
                )
            )

    def _to_cstage(self, defaults=None) -> _ext.CStage:
        """Create a CStage object for the C++ interface.
        This method is protected because, in general, it is better not to
        try to operate directly on the C++ stage object.
        """

        self.validate()
        stage = _ext.CStage()
        if defaults is None:
            defaults = dict()
        stage.timestep = (
            self.solver_timestep
            if self.solver_timestep
            else defaults.get("solver_timestep", 0.1)
        )
        stage.title = self.title
        stage.mode = _ext.CRunMode(int(self.simulation_mode))
        if isinstance(self.save_frequency, str):
            if self.save_frequency == "hourly":
                stage.print_interval = int(np.round(1 / stage.timestep))
            elif self.save_frequency == "daily":
                stage.print_interval = int(np.round(24.0 / stage.timestep))
            elif self.save_frequency == "bystage":
                stage.print_interval = int(
                    np.round(
                        (self.injection_duration + self.rest_duration) / stage.timestep
                    )
                )
            else:
                stage.print_interval = int(self.save_frequency)
        elif isinstance(self.save_frequency, (int, float)):
            stage.print_interval = int(self.save_frequency)
        else:
            stage.print_interval = defaults.get(
                "save_frequency", int(np.round(24.0 / stage.timestep))
            )
        stage.subsequent = 0 if self.set_initial_conditions else 1
        stage.rest_duration = self.rest_duration
        stage.stop_value = (
            int(self.stop_condition) * abs(self.stop_value)
            if self.stop_condition
            else 0
        )
        stage.injection_depth = (
            self.brine_injection_depth if self.brine_injection_depth is not None else 0
        )
        stage.production_depth = (
            self.brine_production_depth
            if self.brine_production_depth is not None
            else 0
        )
        stage.interface_depth = (
            self.brine_interface_depth if self.brine_interface_depth is not None else 0
        )
        stage.injection_rate = self.brine_injection_rate
        stage.inn_tbg_inside_radius = self.inner_tbg_inside_diam / 2.0
        stage.inn_tbg_outside_radius = self.inner_tbg_outside_diam / 2.0
        stage.out_csg_inside_radius = (
            self.outer_csg_inside_diam / 2.0
            if self.outer_csg_inside_diam
            else self.inner_tbg_inside_diam / 2.0
        )
        stage.out_csg_outside_radius = (
            self.outer_csg_outside_diam / 2.0
            if self.outer_csg_outside_diam
            else self.inner_tbg_outside_diam / 2.0
        )
        stage.injection_fluid_sg = self.brine_injection_sg
        stage.cavern_sg = (
            0.0
            if not self.set_initial_conditions
            or self.set_cavern_sg is None
            or self.set_cavern_sg < 1.0
            else self.set_cavern_sg
        )
        stage.timestep = self.solver_timestep
        stage.injection_duration = self.injection_duration
        stage.fill_rate = (
            self.product_injection_rate
            if self.product_injection_rate is not None
            else 0.0
        )
        return stage


@dataclass
class Scenario:
    """A SANSMIC scenario definition used to run a simulation."""

    title: str = None
    """General title for the scenario."""
    comments: str = """"""
    """Comments about the scenario"""
    num_cells: int = None
    """Number of cells to use in the model."""
    geometry_data: Union[List[float], List[List[float]], str] = None
    """The geometry data in a list or list of lists, or a filename."""
    geometry_format: GeometryFormat = GeometryFormat.RADIUS_LIST
    """The format for the geometry data."""
    cavern_height: float = None
    """The total cavern height."""
    floor_depth: float = None
    """The depth from surface datum of the floor."""
    ullage_standoff: float = 40
    """The standoff distance of the lowest brine interface depth."""
    insolubles_ratio: float = 0.04
    """The volume ratio of insoluble material within the salt."""
    units: Units = Units.FT_IN_BBL
    """The units used in describing the scenario."""
    defaults: Dict[str, Union[int, float, str]] = field(default_factory=dict)
    """Default values for a subset of stage attributes, see :attr:`StageDefinition.valid_default_keys`."""
    advanced: AdvancedOptions = field(default_factory=AdvancedOptions)
    """Advanced and/or uncommonly used options"""
    stages: List[StageDefinition] = field(default=None)
    """The activity stages to simulate."""

    def __setattr__(self, name, value):
        """The setattr method is overloaded. IntEnum parameters are
        automatically converted from strings or integers and stages and
        advanced options are converted to the appropriate classes if a
        dictionary is passed."""

        if isinstance(value, str) and value.strip() == "":
            value = None
        if name == "geometry_format" and not isinstance(value, GeometryFormat):
            if isinstance(value, int):
                value = GeometryFormat(value)
            elif isinstance(value, str):
                value = GeometryFormat[
                    value.upper().replace(" ", "_").replace("-", "_")
                ]
            elif isinstance(value, _ext.CGeometryFormat):
                value = GeometryFormat(int(value))
            else:
                TypeError("geometry_format cannot be of type {}".format(type(value)))
        elif name == "units" and not isinstance(value, Units):
            if isinstance(value, int):
                value = Units(value)
            elif isinstance(value, str):
                value = Units[value.upper().replace(" ", "_").replace("-", "_")]
            else:
                TypeError("Units cannot be of type {}".format(type(value)))
        elif (
            name == "advanced"
            and value is not None
            and not isinstance(value, AdvancedOptions)
        ):
            if isinstance(value, dict):
                value = AdvancedOptions.from_dict(value)
            else:
                raise TypeError("advanced cannot be a {}".format(type(value)))
        elif name == "stages":
            if not hasattr(self, "stages") or value is None:
                value = list()
            elif not hasattr(self, name) or self.stages is None:
                converted_stages = list()
                for ct, stage in enumerate(value):
                    if isinstance(stage, dict):
                        if (
                            "set-initial-conditions" in stage
                            or "set_initial_conditions" in stage
                        ):
                            pass
                        else:
                            stage["set-initial-conditions"] = ct == 0
                        stage = StageDefinition.from_dict(stage, defaults=self.defaults)
                    converted_stages.append(stage)
                value = converted_stages
            else:
                TypeError("stages cannot set directly - please use add_stages")
        super().__setattr__(name, value)

    @classmethod
    def from_dict(cls, opts: dict) -> "Scenario":
        """Create a new object from a dictionary of options.

        This method differs from the __init__ constructor by automatically
        converting non-underscore characters - e.g., ``-`` or ``.`` or `` ``
        to underscores and changing keys to lower-case prior to creating
        the object.

        See also: :meth:`~StageDefinition.to_dict`

        Parameters
        ----------
        opts : dict
            The initialization values

        Returns
        -------
        StageDefinition
            the new stage
        """

        new_opts = dict()
        _rename_with_underscore(opts, new_opts)
        return cls(**new_opts)

    def to_dict(self, keep_empty: bool = False):
        """Convert the object's data to a dictionary of options.

        This method differs from the :meth:`~dataclasses.asdict`
        method by automatically converting underscore characters to
        hyphens for a more readable dictionary. Specifically used
        when creating TOML, JSON and YAML files.

        See also: :meth:`~StageDefinition.from_dict`

        Parameters
        ----------
        keep_empty : bool
            Keep values set to None and set them to ``""``

        Returns
        -------
        dict
            the options dictionary
        """

        ret = dict()
        _rename_with_dash(asdict(self), ret)
        ret["advanced"] = self.advanced.to_dict(keep_empty)
        if len(ret["advanced"]) == 0:
            del ret["advanced"]
        del ret["stages"]
        ret["stages"] = list()
        for stage in self.stages:
            stg_dict = stage.to_dict(keep_empty)
            for k in ret["defaults"].keys():
                if (
                    k in stg_dict
                    and k in ret["defaults"]
                    and stg_dict[k] == ret["defaults"][k]
                ):
                    del stg_dict[k]
            ret["stages"].append(stg_dict)
        keys = list(ret.keys())
        for k in keys:
            if ret[k] is None:
                if keep_empty:
                    ret[k] = ""
                else:
                    del ret[k]
        return ret

    def new_stage(self, pos: int = None, **kwargs) -> StageDefinition:
        """Add a new stage in the optionally-specified `pos` position, and create it
        based on keyword arguments. Passes existing :attr:`~Scenario.defaults`
        unless a separate `defaults` dictionary is passed as one of the keyword arguments.

        Parameters
        ----------
        pos : int or None, keyword only
            The position in the stages list to insert the stage,
            by default None which will append to the end.
        kwargs : keyword arguments
            Any valid keyword argument for a :class:`StageDefinition` object

        Returns
        -------
        StageDefinition
            new stage created from the keyword arguments. It will have been
            added into the stage list in the proper position.
        """

        defaults = kwargs.pop("defaults", self.defaults)
        stage = StageDefinition(defaults=defaults, **kwargs)
        if pos is None:
            self.stages.append(stage)
        else:
            self.stages.insert(pos, stage)
        return stage

    def new_simulation(
        self,
        prefix="temp",
        verbosity=0,
        generate_tst_file=True,
        generate_out_file=False,
    ) -> "Simulator":
        """Create a new :class:`Simulator` object.

        Parameters
        ----------
        prefix : str
            The prefix to use when creating output files, by default temp.
        verbosity : int
            A verbosity level to pass to the C++ model, by default 0.
        generate_tst_file : bool, optional
            Generate an old-style .TST file, by default True
        generate_out_file : bool, optional
            Generate an old-style .OUT file, by default False
        """

        return Simulator(self, prefix, verbosity, generate_tst_file, generate_out_file)

    def _to_cscenario(self):
        """Create a C++ model object; in general, this should only be called internally."""

        cscenario = _ext.CScenario()
        cscenario.fraction_insolubles = self.insolubles_ratio
        cscenario.floor_depth = self.floor_depth
        cscenario.geometry_format = _ext.CGeometryFormat(int(self.geometry_format))
        cscenario.num_cells = self.num_cells
        cscenario.ullage_standoff = self.ullage_standoff
        cscenario.cavern_height = self.cavern_height

        # Use defaults for all advanced settings unless they have been changed
        if self.advanced.coallescing_wells is not None:
            cscenario.coallescing_wells = self.advanced.coallescing_wells

        if self.advanced.well_separation is not None:
            cscenario.well_separation = self.advanced.well_separation

        if self.advanced.absolute_error is not None:
            cscenario.absolute_error = self.advanced.absolute_error

        if self.advanced.diffusion_beta is not None:
            cscenario.diffusion_beta = self.advanced.diffusion_beta

        if self.advanced.dissolution_factor is not None:
            cscenario.dissolution_factor = self.advanced.dissolution_factor

        if self.advanced.eddy_coefficient is not None:
            cscenario.eddy_coefficient = self.advanced.eddy_coefficient

        if self.advanced.entrainment_coeff is not None:
            cscenario.entrainment_coeff = self.advanced.entrainment_coeff

        if self.advanced.jet_model_version is not None:
            cscenario.jet_model_version = self.advanced.jet_model_version

        if self.advanced.max_brine_sg is not None:
            cscenario.max_brine_sg = self.advanced.max_brine_sg

        if self.advanced.molecular_diffusion is not None:
            cscenario.molecular_diffusion = self.advanced.molecular_diffusion

        if self.advanced.plume_model_version is not None:
            cscenario.plume_model_version = self.advanced.plume_model_version

        if self.advanced.relative_error is not None:
            cscenario.relative_error = self.advanced.relative_error

        if self.advanced.solid_density is not None:
            cscenario.solid_density = self.advanced.solid_density

        if self.advanced.temperature_model_version is not None:
            cscenario.temperature_model_version = (
                self.advanced.temperature_model_version
            )

        # TODO: This needs to be far more robust
        if self.geometry_format is GeometryFormat.RADIUS_LIST and isinstance(
            self.geometry_data, str
        ):
            radii = list([0])  # Have to pad the data by 1
            with open(self.geometry_data, "r") as fin:
                for line in fin.readlines():
                    if len(line.strip()) > 0:
                        radii.append(float(line.strip()))
            cscenario.radius_vector = radii
        elif isinstance(self.geometry_data, dict):
            if "radii" in self.geometry_data:
                tmp = self.geometry_data["radii"].copy()
                tmp.insert(0, 0)
                cscenario.radius_vector = tmp
            if "depths" in self.geometry_data:
                tmp = self.geometry_data["depths"]
                tmp.insert(0, 0)
                cscenario.depth_vector = tmp
            if "volumes" in self.geometry_data:
                tmp = self.geometry_data["volumes"]
                tmp.insert(0, 0)
                cscenario.volume_vector = tmp
        for stage_num, stage in enumerate(self.stages):
            if stage_num == 0:
                stage.set_initial_conditions = True
            cstage = stage._to_cstage(defaults=self.defaults)
            cscenario.add_stage(cstage)
        return cscenario


class StepwiseIterator:
    """Iterator over each step within an open simulation.

    Examples
    --------
    >>> for step in sim.steps:
    ...     if step % 100 == 0:
    ...         step_res = sim.get_current_state()
    >>> res = sim.results

    """

    def __init__(self, sim: "Simulator"):
        self._sim = sim
        self.__model_stages = self._sim.num_stages
        self.__current_stage = 0
        self.__current_step = 0

    def __iter__(self):
        self.__model_stages = self._sim.num_stages
        self.__current_stage = 0
        self.__current_step = 0
        return self

    def __next__(self):
        self._sim._has_run = True
        if self.__current_stage < self.__model_stages:
            stage_done = self._sim._run_step()
            self.__current_stage = self.__current_stage + stage_done
            self.__current_step = self.__current_step + 1
            return self.__current_stage, self.__current_step
        else:
            raise StopIteration


class Simulator:
    """The SANSMIC Python simulator.

    Parameters
    ----------
    scenario : Scenario
        The scenario to run
    prefix : str, optional
        The output file prefix to use, by default "temp"
    vebosity : int, optional
        The verbosity level for console output, by default 0
    generate_tst_file : bool, optional
        Generate an old-style .TST file, by default True
    generate_out_file : bool, optional
        Generate an old-style .OUT file, by default False
    """

    def __init__(
        self,
        scenario: Union[Scenario, _ext.CModel],
        prefix="temp",
        verbosity=0,
        generate_tst_file=True,
        generate_out_file=False,
    ):
        self._scenario = None
        self._prefix = prefix
        self._cmodel = None
        self._verbosity = verbosity
        self._b_use_tstfile = generate_tst_file
        self._b_use_outfile = generate_out_file
        if isinstance(scenario, Scenario):
            self._scenario = scenario
        else:
            raise TypeError("Invalid scenario")
        self._has_run = False
        self._is_open = False
        self._is_initialized = False
        self._is_finalized = False
        self.__stepwise = None
        self.__results = None

    def __enter__(self):
        self.open(self._prefix)
        self._cmodel.set_verbosity_level(self._verbosity)
        self._cmodel.generate_tst_file(self._b_use_tstfile)
        self._cmodel.generate_out_file(self._b_use_outfile)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    @property
    def steps(self) -> StepwiseIterator:
        """Provides an iterator that will run each step of each stage in turn."""

        if not self._is_open or self._cmodel is None:
            raise RuntimeError("The simulation is not open")
        if self.__stepwise is not None:
            raise RuntimeError("The simulation is already running")
        self._initialize()
        self.__stepwise = StepwiseIterator(self)
        return self.__stepwise

    def __del__(self):
        if self._is_open or self._cmodel is not None:
            try:
                self._cmodel.close_outfiles()
            except Exception as e:
                warnings.warn(e)
        del self._cmodel
        self._cmodel = None

    @property
    def is_open(self) -> bool:
        """Is this simulator attached to a C++ SANSMIC model?"""

        return self._is_open

    @property
    def current_stage(self) -> int:
        """The current stage of the simulation"""

        if not self._is_open or self._cmodel is None:
            raise RuntimeError("The simulation is not open")
        return self._cmodel._get_current_stage()

    @property
    def current_time(self) -> float:
        """The current time (within the stage) of the simulation"""

        if not self._is_open or self._cmodel is None:
            raise RuntimeError("The simulation is not open")
        return self._cmodel._get_current_time()

    @property
    def current_volume(self) -> float:
        """The current simulated cavern volume"""

        if not self._is_open or self._cmodel is None:
            raise RuntimeError("The simulation is not open")
        return self._cmodel._get_current_volume()

    @property
    def is_running(self) -> bool:
        """Is the simulation mid-stage?"""

        if not self._is_open or self._cmodel is None:
            raise RuntimeError("The simulation is not open")
        return self._cmodel.is_running()

    @property
    def num_stages(self) -> int:
        """The total number of stages that have been defined"""

        if not self._is_open or self._cmodel is None:
            raise RuntimeError("The simulation is not open")
        return self._cmodel.num_stages()

    def _initialize(self):
        """Initialize the simulation (for iterators)"""

        if self._is_initialized:
            raise RuntimeError("The simulation has already been initialized")
        self._cmodel.open_outfiles(False)
        self._is_initialized = True
        self._has_run = False
        self._is_finalized = False

    def _run_step(self):
        """Run the next timestep in the specified stage."""

        step_num = self._cmodel.run_next_step()
        return step_num

    def _finalize(self):
        """Finalize the simulation (for iterators)"""

        if self._is_finalized:
            raise RuntimeError("The simulation has already been finalized")
        self._cmodel.close_outfiles()
        self._is_finalized = True
        self._is_initialized = True

    def open(self, prefix=None):
        """Open a simulation.

        Parameters
        ----------
        prefix : str
            The output filename prefix to use, by default None which
            uses the value specified when the Simulator was created.
        """

        if self._is_open or self._cmodel is not None:
            raise RuntimeError("The simulation is already open")
        if prefix is None:
            prefix = self._prefix
        else:
            self._prefix = prefix
        if self._cmodel is None and self._scenario is not None:
            cscenario = self._scenario._to_cscenario()
            self._cmodel = _ext.CModel(prefix)
            self._cmodel.configure(cscenario)
        self.__results = None
        self._is_open = True

    def close(self):
        """Close and garbage collect the C++ model object."""

        if not self._is_finalized and self._is_initialized:
            self._finalize()
        data = self._cmodel._get_results()
        self.__results = Results(data)
        self.__stepwise = None
        del self._cmodel
        self._cmodel = None
        self._is_open = False
        self._is_finalized = False
        self._is_initialized = False

    def run_sim(self):
        """Run the complete simulation; requires the Simulator to have been opened first."""

        if self._cmodel is None:
            raise RuntimeError("The simulation is not open")
        if self.__stepwise is not None:
            raise RuntimeError("The simulation is already running in stepwise mode")
        self._is_finalized = False
        self._is_initialized = True
        self._cmodel.run_sim()
        self._has_run = True
        self._is_initialized = True
        self._is_finalized = True

    def get_current_state(self) -> "Results":
        """Get the current state of the model as a single-timestep results object."""

        if not self._is_open or self._cmodel is None:
            raise RuntimeError("The simulation has not been started")
        data = self._cmodel._get_current_state()
        results = Results(data)
        return results

    @property
    def results(self) -> "Results":
        """The results for an entire simulation."""

        if not self._has_run and self.__results is None:
            warnings.warn(
                "Simulation incomplete - use get_current_state to get partial results"
            )
            return None
        return self.__results


class Results:
    """
    Python container for the results of a sansmic simulation.

    Parameters
    ----------
    data : :class:`~sansmic.libsansmic.CResults`
        results object from the C++ library
    """

    def __init__(self, data: _ext.CResults) -> None:
        self.df_t_1D: pd.DataFrame = None
        """Results that are 1D indexed in the D(time) domain."""
        self.df_z_1D: pd.DataFrame = None
        """Results that are 1D indexed in the D(space) domain."""
        self.df_t_z_2D: pd.DataFrame = None
        """Results that are 2D indexed in the D(time, space) domain."""

        if data is None:
            return
        t = pd.Series(data.t, name="time")
        r = pd.DataFrame(data.r)
        dr = pd.DataFrame(data.dr_0)
        sg = pd.DataFrame(data.sg)
        theta = pd.DataFrame(data.theta)
        Q_inj = pd.DataFrame(data.Q_inj)
        V = pd.DataFrame(data.V)
        f_dis = pd.DataFrame(data.f_dis)
        f_flag = pd.DataFrame(data.f_flag, dtype=int)
        xincl = pd.DataFrame(data.xincl)
        amd = pd.DataFrame(data.amd)
        D_coeff = pd.DataFrame(data.D_coeff)
        dC_dz = pd.DataFrame(data.dC_dz)
        C_old = pd.DataFrame(data.C_old)
        C_new = pd.DataFrame(data.C_new)
        dC_dt = pd.DataFrame(data.dC)
        dr_dt = pd.DataFrame(data.dr)
        C_plm = pd.DataFrame(data.C_plm)
        u_plm = pd.DataFrame(data.u_plm)
        r_plm = pd.DataFrame(data.r_plm)

        self.df_z_1D = pd.DataFrame.from_dict(
            dict(
                z=data.z_0,
                h=data.h_0,
                r=data.r_0,
            )
        )
        self.df_z_1D.index.set_names(["node_idx"], inplace=True)

        self.df_t_1D = pd.DataFrame.from_dict(
            dict(
                t_h=t,
                t_d=t / 24.0,
                step=data.step,
                stage=data.stage,
                phase=data.phase,
                i_inj=data.injCell,
                i_prod=data.prodCell,
                i_plume=data.plumCell,
                i_obi=data.obiCell,
                err_ode=data.err,
                z_inj=data.z_inj,
                z_prod=data.z_prod,
                z_plume=data.z_plm,
                z_obi=data.z_obi,
                z_insol=data.z_insol,
                h_insol=data.h_insol,
                l_jet=data.l_jet,
                u_jet=data.u_jet,
                r_jet=data.r_jet,
                V_inj=data.V_injTot,
                V_fill=data.V_fillTot,
                V_cav=data.V_cavTot,
                V_insol=data.V_insolTot,
                V_vented=data.V_insolVent,
                Q_out=data.Q_out,
                sg_out=data.sg_out,
                sg_ave=data.sg_cavAve,
                dt_h=data.dt,
            ),
        )
        self.df_t_1D.index.set_names(["save_idx"], inplace=True)

        self.df_t_z_2D = pd.DataFrame.from_dict(
            dict(
                r=r.stack(),
                dr=dr.stack(),
                dr_dt=dr_dt.stack(),
                theta=theta.stack(),
                xincl=xincl.stack(),
                V=V.stack(),
                sg=sg.stack(),
                dC_dt=dC_dt.stack(),
                dC_dz=dC_dz.stack(),
                f_flag=f_flag.stack(),
                f_dis=f_dis.stack(),
                D_coeff=D_coeff.stack(),
                r_plm=r_plm.stack(),
                u_plm=u_plm.stack(),
                C_plm=C_plm.stack(),
                Q_inj=Q_inj.stack(),
            )
        )
        self.df_t_z_2D.index.set_names(["save_idx", "node_idx"], inplace=True)

    def __repr__(self) -> str:
        return "<Results: from {:.3g} to {:.3g} d ({:d} stages)>".format(
            self.df_t_1D.t_d.iloc[0],
            self.df_t_1D.t_d.iloc[-1],
            max(self.df_t_1D.stage) - min(self.df_t_1D.stage) + 1,
        )

    @property
    def time(self) -> pd.Series:
        """Time domain: time in days, by timestep"""
        return self.df_t_1D.t_d

    @property
    def injection_cell(self) -> pd.Series:
        """The injection cell, by timestep"""
        return self.df_t_1D.i_inj

    @property
    def production_cell(self) -> pd.Series:
        """The production cell, by timestep"""
        return self.df_t_1D.i_prod

    @property
    def plume_stagnation_cell(self) -> pd.Series:
        """The plume stagnation cell, by timestep"""
        return self.df_t_1D.i_plume

    @property
    def interface_cell(self) -> pd.Series:
        """The brine interface cell, by timestep"""
        return self.df_t_1D.i_obi

    @property
    def ode_convergence(self) -> pd.Series:
        """The ODE convergence factor, by timestep"""
        return self.df_t_1D.err_ode

    @property
    def injection_depth(self) -> pd.Series:
        """The water/brine injection depth, by timestep"""
        return self.df_t_1D.z_inj

    @property
    def production_depth(self) -> pd.Series:
        """The depth of brine production, by timestep"""
        return self.df_t_1D.z_prod

    @property
    def plume_stagnation_depth(self) -> pd.Series:
        """The plume stagnation depth, by timestep"""
        return self.df_t_1D.z_plm

    @property
    def interface_depth(self) -> pd.Series:
        """The brine interface depth, by timestep"""
        return self.df_t_1D.z_obi

    @property
    def cavern_total_depth(self) -> pd.Series:
        """The depth to the new cavern TD, by timestep"""
        return self.df_t_1D.z_insol

    @property
    def insoluble_material_height(self) -> pd.Series:
        """The height of the insoluble materials deposited, by timestep"""
        return self.df_t_1D.h_insol

    @property
    def jet_length(self) -> pd.Series:
        """The length of the injection fluid jet's penetration below EOT, by timestep"""
        return self.df_t_1D.l_jet

    @property
    def jet_velocity(self) -> pd.Series:
        """The velocity of the injection fluid, by timestep"""
        return self.df_t_1D.u_jet

    @property
    def jet_radius(self) -> pd.Series:
        """The radius of the injection fluid jet and starting plume radius, by timestep"""
        return self.df_t_1D.r_jet

    @property
    def cumulative_injection_volume(self) -> pd.Series:
        """Total injected water/brine volume, by timestep"""
        return self.df_t_1D.V_inj

    @property
    def cumulative_fill_volume(self) -> pd.Series:
        """Total product fill volume, by timestep"""
        return self.df_t_1D.V_fill

    @property
    def cavern_volume(self) -> pd.Series:
        """Total cavern volume, by timestep"""
        return self.df_t_1D.V_cav

    @property
    def cumulative_insoluble_material_volume(self) -> pd.Series:
        """Total volume of insoluble materials deposited, by timestep"""
        return self.df_t_1D.V_insol

    @property
    def cumulative_insoluble_material_vented(self) -> pd.Series:
        """Total volume of insoluble materials vented, by timestep"""
        return self.df_t_1D.V_vented

    @property
    def brine_production_rate(self) -> pd.Series:
        """Instantaneous brine production rate, by timestep"""
        return self.df_t_1D.Q_out

    @property
    def brine_production_sg(self) -> pd.Series:
        """Instantaneous brine production cell sg, by timestep"""
        return self.df_t_1D.sg_out

    @property
    def cavern_average_brine_sg(self) -> pd.Series:
        """Instantaneous average cavern brine sg, by timestep"""
        return self.df_t_1D.sg_ave

    @property
    def depths(self) -> pd.Series:
        """Vertical domain: depths, by node"""
        return self.df_z_1D.z

    @property
    def node_heights(self) -> pd.Series:
        """Vertical domain: heights above initial TD, by node"""
        return self.df_z_1D.h

    @property
    def hours(self) -> pd.Series:
        """Time domain: time in hours, by timestep"""
        return self.df_t_1D.t_h

    @property
    def step_size(self) -> pd.Series:
        """Simulation step size, by timestep"""
        return self.df_t_1D.dt

    @property
    def step_number(self) -> pd.Series:
        """Simulation step number, by timestep"""
        return self.df_t_1D.step

    @property
    def stage_number(self) -> pd.Series:
        """Simulation stage number, by timestep"""
        return self.df_t_1D.stage

    @property
    def injection_phase(self) -> pd.Series:
        """Simulation injection(1) or static(0) phase, by timestep"""
        return self.df_t_1D.phase

    @property
    def radius(self) -> pd.DataFrame:
        """Cavern radius, by node and timestep"""
        return self.df_t_z_2D.r.unstack().T

    @property
    def change_in_radius(self) -> pd.DataFrame:
        """Change in cavern radius since previous time, by node and timestep"""
        return self.df_t_z_2D.dr.unstack().T

    @property
    def rate_of_change_in_radius(self) -> pd.DataFrame:
        """Rate of change in cavern radius, by node and timestep"""
        return self.df_t_z_2D.dr_dt.unstack().T

    @property
    def wall_angle(self) -> pd.DataFrame:
        """Dissolution wall angle, by node and timestep"""
        return self.df_t_z_2D.theta.unstack().T

    @property
    def wall_factor(self) -> pd.DataFrame:
        """Dissolution wall factor, by node and timestep"""
        return self.df_t_z_2D.xincl.unstack().T

    @property
    def cell_volume(self) -> pd.DataFrame:
        """Cell volume above node, by node and timestep"""
        return self.df_t_z_2D.V.unstack().T

    @property
    def plume_radius(self) -> pd.DataFrame:
        """Plume radius at node depth, by node and timestep"""
        return self.df_t_z_2D.r_plm.unstack().T

    @property
    def plume_velocity(self) -> pd.DataFrame:
        """Plume velocity at node depth, by node and timestep"""
        return self.df_t_z_2D.u_plm.unstack().T

    @property
    def plume_sg(self) -> pd.DataFrame:
        """Plume concentration(sg) at node depth, by node and timestep"""
        return self.df_t_z_2D.C_plm.unstack().T

    @property
    def cell_injection_rate(self) -> pd.DataFrame:
        """Injection rate within cell above node, by node and timestep"""
        return self.df_t_z_2D.Q_inj.unstack().T

    @property
    def cell_sg(self) -> pd.DataFrame:
        """Concentration(sg) in cell above node, by node and timestep"""
        return self.df_t_z_2D.sg.unstack().T

    @property
    def rate_of_change_in_sg(self) -> pd.DataFrame:
        """Rate of change in concentration in cell above node, by node and timestep"""
        return self.df_t_z_2D.dC_dt.unstack().T

    @property
    def vertical_diffusion_rate(self) -> pd.DataFrame:
        """Vertical change in concentration across node, by node and timestep"""
        return self.df_t_z_2D.dC_dz.unstack().T

    @property
    def state_indicator(self) -> pd.DataFrame:
        """Dissolution state flag, by node and timestep"""
        return self.df_t_z_2D.f_flag.unstack().T

    @property
    def effective_dissolution_factor(self) -> pd.DataFrame:
        """Effective dissolution factor at node, by node and timestep"""
        return self.df_t_z_2D.f_dis.unstack().T

    @property
    def effective_diffusion_coefficient(self) -> pd.DataFrame:
        """Effective diffusion coefficient at node, by node and timestep"""
        return self.df_t_z_2D.D_coeff.unstack().T

    def to_dict(self) -> dict:
        """Convert the results object to a dictionary format."""
        ret = dict()
        for k in ["df_t_1D", "df_z_1D", "df_t_z_2D"]:
            ret[k] = getattr(self, k).to_dict("tight")
        return ret

    @classmethod
    def from_dict(cls, d) -> "Results":
        """Create a new object from a tight dictionary representation."""
        new = cls(None)
        for k in ["df_t_1D", "df_z_1D", "df_t_z_2D"]:
            setattr(new, k, pd.DataFrame.from_dict(d[k], orient="tight"))
        return new

    def to_hdf(
        self,
        filename: str,
        *,
        compression: str = "gzip",
        compression_opts: int = 9,
        shuffle: bool = True,
        fletcher32: bool = True,
        **kwargs,
    ):
        """Write results to an HDF5 file.

        Parameters
        ----------
        filename : str
            A filename with extension ".h5", or filename prefix without extension, to write to

        Keyword Arguments
        -----------------
        compression : str
            The HDF5 compression filter library to use, by default 'gzip'
        compression_opts : int
            The compression level to use, by default 9 (highest)
        shuffle : bool
            Whether to use the shuffle filter, by default True
        fletcher32 : bool
            Whether to use checksum filter, by default True
        kwargs : additional keyword arguments
            Any additional arguments to be passed to the :mod:`h5py` module ``create_dataset`` call.

        """

        if isinstance(h5py, ImportError):
            raise RuntimeError("Optional dependency not installed: h5py") from h5py
        if not filename.lower().endswith(".h5"):
            filename = filename + ".h5"
        with h5py.File(filename, "w") as f:
            f.create_dataset(
                "df_t_1D",
                data=self.df_t_1D.to_records(),
                dtype=[(k, v) for (k, v) in self.df_t_1D.dtypes.items()],
                compression=compression,
                compression_opts=compression_opts,
                shuffle=shuffle,
                fletcher32=fletcher32,
                **kwargs,
            )
            f.create_dataset(
                "df_z_1D",
                data=self.df_z_1D.to_records(),
                dtype=[(k, v) for (k, v) in self.df_z_1D.dtypes.items()],
                compression=compression,
                compression_opts=compression_opts,
                shuffle=shuffle,
                fletcher32=fletcher32,
                **kwargs,
            )
            f.create_dataset(
                "df_t_z_2D",
                data=self.df_t_z_2D.to_records(),
                dtype=[(k, v) for (k, v) in self.df_t_z_2D.dtypes.items()],
                compression=compression,
                compression_opts=compression_opts,
                shuffle=shuffle,
                fletcher32=fletcher32,
                **kwargs,
            )

    @classmethod
    def from_hdf(cls, filename: str) -> "Results":
        """Read results from an HDF5 file.

        Parameters
        ----------
        filename : str
            A filename with extension ".h5", or filename prefix without extension, to read from.

        """

        results = cls(None)
        if isinstance(h5py, ImportError):
            raise RuntimeError("Optional dependency not installed: h5py") from h5py
        if not filename.lower().endswith(".h5"):
            filename = filename + ".h5"
        with h5py.File(filename, "r") as f:
            results.df_t_1D = pd.DataFrame(
                np.array(f["df_t_1D"], dtype=f["df_t_1D"].dtype).view(np.recarray)
            )
            results.df_z_1D = pd.DataFrame(
                np.array(f["df_z_1D"], dtype=f["df_z_1D"].dtype).view(np.recarray)
            )
            results.df_t_z_2D = pd.DataFrame(
                np.array(f["df_t_z_2D"], dtype=f["df_t_z_2D"].dtype).view(np.recarray)
            )
        return results


class _OutDataBlock:
    """Useful for processing old SANSMIC .out files"""

    def __init__(self):
        self.t = None
        self.dt = None
        self.t_prime = None
        self.phase = None
        self.stage = None
        self.injCell = None
        self.prodCell = None
        self.obiCell = None
        self.plumCell = None
        self.z_inj = None
        self.z_prod = None
        self.z_obi = None
        self.z_plm = None
        self.l_jet = None
        self.r_jet = None
        self.u_jet = None
        self.idx = list()
        self.h = list()
        self.r = list()
        self.dr_0 = list()
        self.sg = list()
        self.theta = list()
        self.Q_inj = list()
        self.V = list()
        self.f_dis = list()
        self.f_flag = list()
        self.xincl = list()
        self.amd = list()
        self.D_coeff = list()
        self.dC_dz = list()
        self.C_old = list()
        self.C_new = list()
        self.dC = list()
        self.dr = list()
        self.C_plm = list()
        self.u_plm = list()
        self.r_plm = list()
        self.h_insol = None
        self.V_cavTot = None
        self.Q_out = None
        self.sg_out = None
        self.V_insolTot = None
        self.z_insol = None
        self.z_obi = None
        self.V_insolVent = None
        self.V_brine = None
        self.V_ullage = None
        self.V_usable = None
        self.err = None
        self.step = None


class _OutputData(_OutDataBlock):
    """Useful for processing old SANSMIC .out files"""

    def __init__(self):
        super().__init__()
        self.timeTotal = None
        self.r_0 = None
        self.z_0 = None
        self.h_0 = None
        self.V_injTot = None
        self.V_fillTot = None
        self.sg_cavAve = None



# coding: utf-8

import os
import unittest
from os.path import abspath, dirname, join

import numpy as np
import pandas as pd
import sansmic

testdir = dirname(abspath(str(__file__)))


class TestRegression(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.nameF = "baseline"
        cls.datF = join(testdir, cls.nameF + ".dat")
        cls.tstF = sansmic.io.read_tst_file(join(testdir, cls.nameF))
        cls.resF = sansmic.io.read_classic_out_ddl(join(testdir, cls.nameF))
        cls.namePy = "regression"

    def runTest(self):
        model = sansmic.io.read_scenario(self.datF)
        sansmic.io.write_scenario(model, join(testdir, self.namePy + ".toml"))
        with model.new_simulation(join(testdir, self.namePy)) as sim:
            sim.run_sim()
        resPy = sim.results
        tstPy = sansmic.io.read_tst_file(join(testdir, self.namePy))
        Stats = sansmic.model.pd.DataFrame.from_dict(
            dict(
                rmse=np.sqrt(np.mean((self.tstF - tstPy) ** 2, 0)),
                max_rel=((self.tstF - tstPy).abs() / abs(self.tstF)).max(),
                rmse_rel_err=(np.sqrt(np.mean((self.tstF - tstPy) ** 2, 0)))
                / self.tstF.abs().max(),
                min_F=abs(self.tstF).min(),
                min_Py=abs(tstPy).min(),
                max_F=abs(self.tstF).max(),
                max_Py=abs(tstPy).max(),
            )
        )
        self.assertLessEqual(Stats.rmse_rel_err["t_d"], 1.0e-4)
        self.assertLessEqual(Stats.rmse_rel_err["V_cav"], 5.0e-4)
        self.assertLessEqual(Stats.rmse_rel_err["err_ode"], 5.0e-4)
        self.assertLessEqual(Stats.rmse_rel_err["sg_out"], 5.0e-4)
        self.assertLessEqual(Stats.rmse_rel_err["sg_ave"], 5.0e-4)
        self.assertLessEqual(Stats.rmse_rel_err["V_insol"], 1.0e-3)
        self.assertLessEqual(Stats.rmse_rel_err["z_insol"], 5.0e-4)
        self.assertLessEqual(Stats.rmse_rel_err["z_obi"], 5.0e-4)
        self.assertLessEqual(Stats.rmse_rel_err["V_vented"], 1.0e-2)
        self.assertLessEqual(Stats.rmse_rel_err["Q_inj"], 5.0e-4)
        self.assertLessEqual(Stats.rmse_rel_err["Q_fill"], 5.0e-4)
        self.assertLessEqual(Stats.rmse_rel_err["V_inj"], 1.0e-3)
        self.assertLessEqual(Stats.rmse_rel_err["V_fill"], 1.0e-3)



# coding: utf-8

import os
import unittest
from os.path import abspath, dirname, join

import numpy as np
import pandas as pd
import sansmic, sansmic.io, sansmic.model
import tempfile

testdir = dirname(abspath(str(__file__)))


class TestReadAndWriteScenarios(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        # Set up temporary directory
        cls.tempdir = tempfile.TemporaryDirectory()
        cls.tempdirname = cls.tempdir.name
        # cls.tempdirname = abspath('tests')

        # Set temporary filenames for the static data
        cls.withdrawal_dat = os.path.join(cls.tempdirname, "withdrawal.dat")
        cls.ordinary_dir_dat = os.path.join(cls.tempdirname, "ordinary-dir.dat")
        cls.ordinary_rev_dat = os.path.join(cls.tempdirname, "ordinary-rev.dat")
        cls.leach_fill_dat = os.path.join(cls.tempdirname, "leach-fill.txt")

        # Write the .DAT formatted data
        with open(cls.withdrawal_dat, "w") as fin:
            fin.write(cls._WITHDRAWAL_DAT_DATA)
        with open(cls.ordinary_dir_dat, "w") as fin:
            fin.write(cls._ORDINARY_DIRECT_DATA)
        with open(cls.ordinary_rev_dat, "w") as fin:
            fin.write(cls._ORDINARY_REVERSE_DATA)
        with open(cls.leach_fill_dat, "w") as fin:
            fin.write(cls._LEACH_FILL_DATA)

        # Set temporary filenames
        cls.withdrawal_toml = os.path.join(cls.tempdirname, "withdrawal.toml")
        cls.ordinary_dir_json = os.path.join(cls.tempdirname, "ordinary-dir.json")
        cls.ordinary_rev_yaml = os.path.join(cls.tempdirname, "ordinary-rev.yaml")
        cls.leach_fill_format = os.path.join(cls.tempdirname, "leach-fill-new.txt")

    def test_read_scenario_leachfill(self):
        scenario = sansmic.io.read_scenario(self.leach_fill_dat, format="dat")
        self.assertEqual(len(scenario.stages), 1)
        self.assertEqual(scenario.stages[0].title, "Leach fill")
        self.assertEqual(scenario.num_cells, 100)
        self.assertEqual(scenario.stages[0].simulation_mode, 2)
        self.assertIs(
            scenario.stages[0].simulation_mode, sansmic.model.SimulationMode.LEACH_FILL
        )
        self.assertEqual(scenario.stages[0].save_frequency, 240)
        self.assertTrue(scenario.stages[0].set_initial_conditions)
        self.assertEqual(scenario.stages[0].rest_duration, 1440)
        self.assertIn(scenario.advanced.coallescing_wells, [1, None])
        self.assertEqual(scenario.geometry_format, 0)
        self.assertIs(
            scenario.geometry_format, sansmic.model.GeometryFormat.RADIUS_LIST
        )
        self.assertEqual(scenario.stages[0].stop_value, -200.0)
        self.assertIs(
            scenario.stages[0].stop_condition, sansmic.model.StopCondition.DEPTH
        )
        self.assertEqual(scenario.cavern_height, 1000.0)
        self.assertEqual(scenario.stages[0].brine_injection_depth, 32.0)
        self.assertEqual(scenario.stages[0].brine_production_depth, 350.0)
        self.assertEqual(scenario.stages[0].brine_interface_depth, 900.0)
        self.assertEqual(scenario.ullage_standoff, 20.0)
        self.assertEqual(scenario.stages[0].brine_injection_rate, 240000.0)
        self.assertEqual(scenario.stages[0].inner_tbg_inside_diam / 2, 4.925)
        self.assertEqual(scenario.stages[0].inner_tbg_outside_diam / 2, 5.375)
        self.assertEqual(scenario.stages[0].outer_csg_inside_diam / 2, 4.925)
        self.assertEqual(scenario.stages[0].outer_csg_outside_diam / 2, 5.375)
        self.assertEqual(scenario.stages[0].brine_injection_sg, 1.03)
        self.assertEqual(scenario.stages[0].set_cavern_sg, 1.2019)
        self.assertEqual(scenario.stages[0].solver_timestep, 0.1)
        self.assertEqual(scenario.stages[0].injection_duration, 72)
        self.assertEqual(scenario.stages[0].product_injection_rate, 240000.0)
        self.assertIsInstance(scenario.geometry_data, dict)
        self.assertIn("radii", scenario.geometry_data)
        self.assertEqual(len(scenario.geometry_data["radii"]), 101)
        self.assertEqual(scenario.floor_depth, 4000.0)
        self.assertEqual(scenario.advanced.dissolution_factor, 1.01)
        self.assertEqual(scenario.insolubles_ratio, 0.05)

    def test_read_scenario_withdrawal(self):
        scenario = sansmic.io.read_scenario(self.withdrawal_dat)
        self.assertEqual(len(scenario.stages), 1)
        self.assertEqual(scenario.stages[0].title, "Withdrawal leach")
        self.assertEqual(scenario.num_cells, 100)
        self.assertEqual(scenario.stages[0].simulation_mode, 1)
        self.assertIs(
            scenario.stages[0].simulation_mode, sansmic.model.SimulationMode.WITHDRAWAL
        )
        self.assertEqual(scenario.stages[0].save_frequency, 240)
        self.assertTrue(scenario.stages[0].set_initial_conditions)
        self.assertEqual(scenario.stages[0].rest_duration, 1440)
        self.assertIn(scenario.advanced.coallescing_wells, [1, None])
        self.assertEqual(scenario.geometry_format, 0)
        self.assertIs(
            scenario.geometry_format, sansmic.model.GeometryFormat.RADIUS_LIST
        )
        self.assertEqual(scenario.stages[0].stop_value, 0.0)
        self.assertIs(
            scenario.stages[0].stop_condition, sansmic.model.StopCondition.DURATION
        )
        self.assertEqual(scenario.cavern_height, 1000.0)
        self.assertEqual(scenario.stages[0].brine_injection_depth, 32.0)
        self.assertEqual(scenario.stages[0].brine_production_depth, 1000.0)
        self.assertEqual(scenario.stages[0].brine_interface_depth, 100.0)
        self.assertEqual(scenario.ullage_standoff, 20.0)
        self.assertEqual(scenario.stages[0].brine_injection_rate, 240000.0)
        self.assertEqual(scenario.stages[0].inner_tbg_inside_diam / 2, 4.925)
        self.assertEqual(scenario.stages[0].inner_tbg_outside_diam / 2, 5.375)
        self.assertEqual(scenario.stages[0].outer_csg_inside_diam / 2, 4.925)
        self.assertEqual(scenario.stages[0].outer_csg_outside_diam / 2, 5.375)
        self.assertEqual(scenario.stages[0].brine_injection_sg, 1.03)
        self.assertEqual(scenario.stages[0].set_cavern_sg, 1.2019)
        self.assertEqual(scenario.stages[0].solver_timestep, 0.1)
        self.assertEqual(scenario.stages[0].injection_duration, 72)
        self.assertEqual(scenario.stages[0].product_injection_rate, 0.0)
        self.assertIsInstance(scenario.geometry_data, dict)
        self.assertIn("radii", scenario.geometry_data)
        self.assertEqual(len(scenario.geometry_data["radii"]), 101)
        self.assertEqual(scenario.floor_depth, 4000.0)
        self.assertIn(scenario.advanced.dissolution_factor, [1.0, None])
        self.assertEqual(scenario.insolubles_ratio, 0.05)

    def test_read_scenario_ordinary_direct(self):
        scenario = sansmic.io.read_scenario(self.ordinary_dir_dat)
        self.assertEqual(len(scenario.stages), 1)
        self.assertEqual(scenario.stages[0].title, "Ordinary leach - direct")
        self.assertEqual(scenario.num_cells, 100)
        self.assertEqual(scenario.stages[0].simulation_mode, 0)
        self.assertIs(
            scenario.stages[0].simulation_mode, sansmic.model.SimulationMode.ORDINARY
        )
        self.assertEqual(scenario.stages[0].save_frequency, 240)
        self.assertTrue(scenario.stages[0].set_initial_conditions)
        self.assertEqual(scenario.stages[0].rest_duration, 1440)
        self.assertIn(scenario.advanced.coallescing_wells, [1, None])
        self.assertEqual(scenario.geometry_format, 0)
        self.assertIs(
            scenario.geometry_format, sansmic.model.GeometryFormat.RADIUS_LIST
        )
        self.assertEqual(scenario.stages[0].stop_value, 4000000.0)
        self.assertIs(
            scenario.stages[0].stop_condition, sansmic.model.StopCondition.VOLUME
        )
        self.assertEqual(scenario.cavern_height, 1000.0)
        self.assertEqual(scenario.stages[0].brine_injection_depth, 100.0)
        self.assertEqual(scenario.stages[0].brine_production_depth, 800.0)
        self.assertEqual(scenario.stages[0].brine_interface_depth, 900.0)
        self.assertEqual(scenario.ullage_standoff, 20.0)
        self.assertEqual(scenario.stages[0].brine_injection_rate, 240000.0)
        self.assertEqual(scenario.stages[0].inner_tbg_inside_diam / 2, 4.925)
        self.assertEqual(scenario.stages[0].inner_tbg_outside_diam / 2, 5.375)
        self.assertEqual(scenario.stages[0].outer_csg_inside_diam / 2, 6.925)
        self.assertEqual(scenario.stages[0].outer_csg_outside_diam / 2, 7.375)
        self.assertEqual(scenario.stages[0].brine_injection_sg, 1.1)
        self.assertEqual(scenario.stages[0].set_cavern_sg, 1.2019)
        self.assertEqual(scenario.stages[0].solver_timestep, 0.1)
        self.assertEqual(scenario.stages[0].injection_duration, 72)
        self.assertEqual(scenario.stages[0].product_injection_rate, 0.0)
        self.assertIsInstance(scenario.geometry_data, dict)
        self.assertIn("radii", scenario.geometry_data)
        self.assertEqual(len(scenario.geometry_data["radii"]), 101)
        self.assertEqual(scenario.floor_depth, 4000.0)
        self.assertIn(scenario.advanced.dissolution_factor, [1.0, None])
        self.assertEqual(scenario.insolubles_ratio, 0.05)

    def test_read_scenario_ordinary_reverse_multistage(self):
        scenario = sansmic.io.read_scenario(self.ordinary_rev_dat)
        self.assertEqual(len(scenario.stages), 3)
        self.assertEqual(scenario.stages[0].title, "Ordinary leach stage 1")
        self.assertEqual(scenario.num_cells, 100)
        self.assertEqual(scenario.stages[0].simulation_mode, 0)
        self.assertIs(
            scenario.stages[0].simulation_mode, sansmic.model.SimulationMode.ORDINARY
        )
        self.assertEqual(scenario.stages[0].save_frequency, 240)
        self.assertTrue(scenario.stages[0].set_initial_conditions)
        self.assertEqual(scenario.stages[0].rest_duration, 1440)
        self.assertIn(scenario.advanced.coallescing_wells, [1, None])
        self.assertEqual(scenario.geometry_format, 0)
        self.assertIs(
            scenario.geometry_format, sansmic.model.GeometryFormat.RADIUS_LIST
        )
        self.assertEqual(scenario.stages[0].stop_value, 0.0)
        self.assertIs(
            scenario.stages[0].stop_condition, sansmic.model.StopCondition.DURATION
        )
        self.assertEqual(scenario.cavern_height, 1000.0)
        self.assertEqual(scenario.stages[0].brine_injection_depth, 800.0)
        self.assertEqual(scenario.stages[0].brine_production_depth, 100.0)
        self.assertEqual(scenario.stages[0].brine_interface_depth, 900.0)
        self.assertEqual(scenario.ullage_standoff, 20.0)
        self.assertEqual(scenario.stages[0].brine_injection_rate, 240000.0)
        self.assertEqual(scenario.stages[0].inner_tbg_inside_diam / 2, 4.925)
        self.assertEqual(scenario.stages[0].inner_tbg_outside_diam / 2, 5.375)
        self.assertEqual(scenario.stages[0].outer_csg_inside_diam / 2, 4.925)
        self.assertEqual(scenario.stages[0].outer_csg_outside_diam / 2, 5.375)
        self.assertEqual(scenario.stages[0].brine_injection_sg, 1.03)
        self.assertEqual(scenario.stages[0].set_cavern_sg, 1.2019)
        self.assertEqual(scenario.stages[0].solver_timestep, 0.1)
        self.assertEqual(scenario.stages[0].injection_duration, 72)
        self.assertEqual(scenario.stages[0].product_injection_rate, 0.0)
        self.assertIsInstance(scenario.geometry_data, dict)
        self.assertIn("radii", scenario.geometry_data)
        self.assertEqual(len(scenario.geometry_data["radii"]), 101)
        self.assertEqual(scenario.floor_depth, 4000.0)
        self.assertIn(scenario.advanced.dissolution_factor, [1.0, None])
        self.assertEqual(scenario.insolubles_ratio, 0.05)
        self.assertEqual(scenario.stages[1].title, "Ordinary leach stage 2")
        self.assertEqual(scenario.stages[1].simulation_mode, 0)
        self.assertIs(
            scenario.stages[1].simulation_mode, sansmic.model.SimulationMode.ORDINARY
        )
        self.assertEqual(scenario.stages[1].save_frequency, 240)
        self.assertFalse(scenario.stages[1].set_initial_conditions)
        self.assertEqual(scenario.stages[1].rest_duration, 1440)
        self.assertEqual(scenario.stages[1].stop_value, 0.0)
        self.assertIs(
            scenario.stages[1].stop_condition, sansmic.model.StopCondition.DURATION
        )
        self.assertEqual(scenario.stages[1].brine_injection_depth, 800.0)
        self.assertEqual(scenario.stages[1].brine_production_depth, 100.0)
        self.assertEqual(scenario.stages[1].brine_interface_depth, 0.0)
        self.assertEqual(scenario.stages[1].brine_injection_rate, 240000.0)
        self.assertEqual(scenario.stages[1].inner_tbg_inside_diam / 2, 4.925)
        self.assertEqual(scenario.stages[1].inner_tbg_outside_diam / 2, 5.375)
        self.assertEqual(scenario.stages[1].outer_csg_inside_diam / 2, 6.925)
        self.assertEqual(scenario.stages[1].outer_csg_outside_diam / 2, 7.375)
        self.assertEqual(scenario.stages[1].brine_injection_sg, 1.03)
        self.assertIsNone(scenario.stages[1].set_cavern_sg)
        self.assertEqual(scenario.stages[1].solver_timestep, 0.1)
        self.assertEqual(scenario.stages[1].injection_duration, 72)
        self.assertEqual(scenario.stages[1].product_injection_rate, 0.0)
        self.assertEqual(scenario.stages[2].title, "Ordinary leach stage 3")
        self.assertEqual(scenario.stages[2].simulation_mode, 0)
        self.assertIs(
            scenario.stages[2].simulation_mode, sansmic.model.SimulationMode.ORDINARY
        )
        self.assertEqual(scenario.stages[2].save_frequency, 240)
        self.assertFalse(scenario.stages[2].set_initial_conditions)
        self.assertEqual(scenario.stages[2].rest_duration, 1440)
        self.assertEqual(scenario.stages[2].stop_value, 0.0)
        self.assertIs(
            scenario.stages[2].stop_condition, sansmic.model.StopCondition.DURATION
        )
        self.assertEqual(scenario.stages[2].brine_injection_depth, 800.0)
        self.assertEqual(scenario.stages[2].brine_production_depth, 100.0)
        self.assertEqual(scenario.stages[2].brine_interface_depth, 0.0)
        self.assertEqual(scenario.stages[2].brine_injection_rate, 240000.0)
        self.assertEqual(scenario.stages[2].inner_tbg_inside_diam / 2, 4.925)
        self.assertEqual(scenario.stages[2].inner_tbg_outside_diam / 2, 5.375)
        self.assertEqual(scenario.stages[2].outer_csg_inside_diam / 2, 8.925)
        self.assertEqual(scenario.stages[2].outer_csg_outside_diam / 2, 9.375)
        self.assertEqual(scenario.stages[2].brine_injection_sg, 1.03)
        self.assertIsNone(scenario.stages[2].set_cavern_sg)
        self.assertEqual(scenario.stages[2].solver_timestep, 0.1)
        self.assertEqual(scenario.stages[2].injection_duration, 72)
        self.assertEqual(scenario.stages[2].product_injection_rate, 0.0)

    def test_roundtrip_toml(self):
        scenario = sansmic.io.read_scenario(self.withdrawal_dat)
        scenario.title = ""
        sansmic.io.write_scenario(scenario, self.withdrawal_toml)
        scenario2 = sansmic.io.read_scenario(self.withdrawal_toml)
        scenario2.title = ""
        self.assertEqual(scenario, scenario2)

    def test_roundtrip_yaml(self):
        self.maxDiff = None
        scenario = sansmic.io.read_scenario(self.ordinary_rev_dat)
        scenario.title = ""
        scenario.stages[1].brine_interface_depth = 0
        scenario.stages[2].brine_interface_depth = 0
        scenario.stages[1].set_initial_conditions = None
        scenario.stages[2].set_initial_conditions = None

        sansmic.io.write_scenario(scenario, self.ordinary_rev_yaml, redundant=True)
        scenario2 = sansmic.io.read_scenario(self.ordinary_rev_yaml)
        scenario2.title = ""
        self.assertDictEqual(
            scenario.to_dict(keep_empty=True), scenario2.to_dict(keep_empty=True)
        )

    def test_roundtrip_json(self):
        scenario = sansmic.io.read_scenario(self.ordinary_dir_dat)
        scenario.title = ""
        sansmic.io.write_scenario(scenario, self.ordinary_dir_json)
        scenario2 = sansmic.io.read_scenario(self.ordinary_dir_json)
        scenario2.title = ""
        self.assertEqual(scenario, scenario2)

    def test_roundtrip_spec_format(self):
        scenario = sansmic.io.read_scenario(self.leach_fill_dat, format="dat")
        scenario.title = ""

        sansmic.io.write_scenario(scenario, self.leach_fill_format, format="toml")
        scenario2 = sansmic.io.read_scenario(self.leach_fill_format, format="toml")
        scenario2.title = ""
        self.assertEqual(scenario, scenario2)

        sansmic.io.write_scenario(scenario, self.leach_fill_format, format="json")
        scenario3 = sansmic.io.read_scenario(self.leach_fill_format, format="json")
        scenario3.title = ""
        self.assertEqual(scenario, scenario3)

        sansmic.io.write_scenario(scenario, self.leach_fill_format, format="yaml")
        scenario4 = sansmic.io.read_scenario(self.leach_fill_format, format="yaml")
        scenario4.title = ""
        self.assertEqual(scenario, scenario4)

    @classmethod
    def tearDownClass(cls):
        cls.tempdir.cleanup()

    _WITHDRAWAL_DAT_DATA = """Withdrawal leach
100 1   240 0   0   1440    1   0   0
1000.   32.0    1000.0  100.0   20.0
240000.0
4.925   5.3750  4.925   5.3750
1.03    1.2019
0.1     72
0.0 0.0 0.0
10
50
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
90
80
50
33
10
4
2
2
2
2
2
1.0   0.05   4000.   4000.
END
"""

    _ORDINARY_DIRECT_DATA = """Ordinary leach - direct
100	0	240	0	0	1440	 1	0	4000000.0
1000.   100.0  800.0	900.0   20.0
240000.0
4.925	5.3750	6.925	7.3750
1.1 1.2019
0.1     72
0.0	0.0	0.0
10
50
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
90
80
50
33
10
4
2
2
2
2
2
1.0   0.05   4000.   4000.
END
"""

    _ORDINARY_REVERSE_DATA = """Ordinary leach stage 1
100	0	240	0	0	1440	 1	0	0
1000.   800.0  100.0	900.0   20.0
240000.0
4.925	5.3750	4.925	5.3750
1.03	1.2019
0.1     72
0.0	0.0	0.0
10
50
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
90
80
50
33
10
4
2
2
2
2
2
1.0   0.05   4000.   4000.
Ordinary leach stage 2
100	0	240	1	0	1440	 1	0	0
1000.   800.0  100.0	0.0   20.0
240000.0
4.925	5.3750	6.925	7.3750
1.03	0.0
0.1     72
0.0	0.0	0.0
Ordinary leach stage 3
100	0	240	1	0	1440	 1	0	0
1000.   800.0  100.0	0.0   20.0
240000.0
4.925	5.3750	8.925	9.3750
1.03	1.2019
0.1     72
0.0	0.0	0.0
END
"""

    _LEACH_FILL_DATA = """Leach fill
100	2	240	0	0	1440	 1	0	-200.0
1000.   32.0	350.0  900.0   20.0
240000.0
4.925	5.3750	4.925	5.3750
1.03	1.2019
0.1     72
240000.0	0.0	0.0
10
50
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
90
80
50
33
10
4
2
2
2
2
2
1.01   0.05   4000.   4000.
END
"""


if __name__ == "__main__":
    unittest.main()



# coding: utf-8

import glob
import os
import tempfile
import unittest
import subprocess
from os.path import abspath, dirname, join

import numpy as np
import pandas as pd
import sansmic
import sansmic.app
import sansmic.io
import sansmic.model

testdir = dirname(abspath(str(__file__)))


class TestApplication(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        # Set up temporary directory
        cls.tempdir = tempfile.TemporaryDirectory()
        cls.tempdirname = cls.tempdir.name
        # cls.tempdirname = abspath('.')

        # Set temporary filenames
        cls.withdrawal_dat = os.path.join(cls.tempdirname, "withdrawal.dat")
        cls.withdrawal_toml = os.path.join(cls.tempdirname, "withdrawal.toml")

        # Write the .DAT formatted data
        with open(cls.withdrawal_dat, "w") as fin:
            fin.write(cls._WITHDRAWAL_DAT)

    def test_run_app(self):
        """Test the 'sansmic' command line application."""
        scenario = sansmic.io.read_scenario(self.withdrawal_dat)
        with scenario.new_simulation() as sim:
            sim.run_sim()
        res0 = sim.results
        res1 = sansmic.app.main(
            args=[
                self.withdrawal_dat,
                "--no-toml",
                "--no-csv",
                "--no-json",
                "--no-hdf",
                "--no-tst",
                "--no-old-out",
            ],
            ret=True,
        )
        res2 = sansmic.app.main(
            args=[
                self.withdrawal_dat,
                "-o",
                join(self.tempdirname, "app-test"),
                "--toml",
                "--csv",
                "--json",
                "--hdf",
                "--tst",
                "--old-out",
            ],
            ret=False,
        )
        res3 = sansmic.app.main(
            args=[
                self.withdrawal_dat,
                "--no-toml",
                "--no-csv",
                "--no-json",
                "--no-hdf",
                "--no-tst",
                "--no-old-out",
                "-v",
            ],
            ret=True,
        )
        res4 = sansmic.app.main(
            args=[
                self.withdrawal_dat,
                "--no-toml",
                "--no-csv",
                "--no-json",
                "--no-hdf",
                "--no-tst",
                "--no-old-out",
                "-v",
                "-v",
            ],
            ret=True,
        )
        res5 = sansmic.app.main(
            args=[
                self.withdrawal_dat,
                "--no-toml",
                "--no-csv",
                "--no-json",
                "--no-hdf",
                "--no-tst",
                "--no-old-out",
                "-q",
            ],
            ret=True,
        )

        res6 = sansmic.app.main(
            args=[
                self.withdrawal_dat,
                "--debug",
            ],
            ret=True,
        )

        self.assertIsNotNone(res1)
        self.assertIsNone(res2)

        self.assertEqual(len(glob.glob(join(self.tempdirname, "app-test.toml"))), 1)
        self.assertEqual(len(glob.glob(join(self.tempdirname, "app-test.log"))), 1)
        self.assertEqual(len(glob.glob(join(self.tempdirname, "app-test*.csv"))), 7)
        self.assertEqual(len(glob.glob(join(self.tempdirname, "app-test.json"))), 1)
        self.assertEqual(len(glob.glob(join(self.tempdirname, "app-test.h5"))), 1)
        self.assertEqual(len(glob.glob(join(self.tempdirname, "app-test.tst"))), 1)
        self.assertEqual(len(glob.glob(join(self.tempdirname, "app-test.out"))), 1)
        self.assertEqual(len(glob.glob(join(self.tempdirname, "app-test*"))), 13)

        self.assertTrue((res0.df_t_1D == res1.df_t_1D).all().all())
        self.assertTrue((res0.df_z_1D == res1.df_z_1D).all().all())
        self.assertTrue((res0.df_t_z_2D == res1.df_t_z_2D).all().all())

        self.assertTrue((res0.df_t_1D == res3.df_t_1D).all().all())
        self.assertTrue((res0.df_z_1D == res3.df_z_1D).all().all())
        self.assertTrue((res0.df_t_z_2D == res3.df_t_z_2D).all().all())

        self.assertTrue((res0.df_t_1D == res4.df_t_1D).all().all())
        self.assertTrue((res0.df_z_1D == res4.df_z_1D).all().all())
        self.assertTrue((res0.df_t_z_2D == res4.df_t_z_2D).all().all())

        self.assertTrue((res0.df_t_1D == res5.df_t_1D).all().all())
        self.assertTrue((res0.df_z_1D == res5.df_z_1D).all().all())
        self.assertTrue((res0.df_t_z_2D == res5.df_t_z_2D).all().all())

        self.assertTrue((res0.df_t_1D == res6.df_t_1D).all().all())
        self.assertTrue((res0.df_z_1D == res6.df_z_1D).all().all())
        self.assertTrue((res0.df_t_z_2D == res6.df_t_z_2D).all().all())

        with self.assertRaises(FileNotFoundError):
            sansmic.app.main(
                args=[
                    "thisWontWork.TOML",
                ],
            )

    def test_convert_app(self):
        """Test the 'sansmic-convert' command line application."""
        scenario0 = sansmic.io.read_scenario(self.withdrawal_dat)
        scenario0.title = ""
        sansmic.app.convert(args=[self.withdrawal_dat, self.withdrawal_toml])
        scenario1 = sansmic.io.read_scenario(self.withdrawal_toml)
        scenario1.title = ""
        self.assertEqual(scenario0, scenario1)
        sansmic.app.convert(args=[self.withdrawal_dat, self.withdrawal_toml, "--full"])
        scenario2 = sansmic.io.read_scenario(self.withdrawal_toml)
        scenario2.title = ""
        self.maxDiff = None
        self.assertDictEqual(scenario0.to_dict(), scenario2.to_dict())

        with self.assertRaises(FileNotFoundError):
            sansmic.app.convert(
                args=["thisWontWork.TOML", "neither.will.this."],
            )

    def test_sansmic_cmd(self):
        proc = subprocess.run(
            [
                "sansmic",
                self.withdrawal_dat,
                "-o",
                join(self.tempdirname, "sansmic-call-test"),
                "-v",
            ],
            capture_output=True,
            # avoid having to explicitly encode
            text=True,
        )
        data = proc.stdout
        result = proc.returncode
        self.assertTrue("Progress..." in data)
        self.assertTrue("960000.0" in data)
        self.assertEqual(
            len(glob.glob(join(self.tempdirname, "sansmic-call-test*"))), 9
        )

    @classmethod
    def tearDownClass(cls):
        cls.tempdir.cleanup()

    _WITHDRAWAL_DAT = """Withdrawal leach
100 1   240 0   0   1440    1   0   0
1000.   32.0    1000.0  100.0   20.0
240000.0
4.925   5.3750  4.925   5.3750
1.03    1.2019
0.1     72
0.0 0.0 0.0
10
50
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
100
90
80
50
33
10
4
2
2
2
2
2
1.0   0.05   4000.   4000.
Stage 2
100 1   240 0   0   120    1   0   0
1000.   32.0    1000.0  100.0   20.0
240000.0
4.925   5.3750  4.925   5.3750
1.03    1.2019
0.1     24
0.0 0.0 0.0
END
"""



"""Setup SANSMIC"""

# Always prefer setuptools over distutils
from os import path
from glob import glob

from setuptools import find_packages, setup
from pybind11.setup_helpers import Pybind11Extension, build_ext

__version__ = "1.0.0-rc.1"

here = path.abspath(path.dirname(__file__))

ext_modules = [
    Pybind11Extension(
        "sansmic.libsansmic",
        sources=sorted(glob("src/ext_modules/libsansmic/*.cpp")),
        cxx_std=17,
        # Example: passing in the version to the compiled code
        define_macros=[
            ("VERSION_INFO", '"{}"'.format(__version__)),
        ],
    ),
]

setup(
    name="sansmic",  # Required
    version=__version__,  # Required
    description="The SANSMIC solution mining code",  # Optional
    # long_description=long_description,  # Optional
    # long_description_content_type='text/markdown',  # Optional (see note above)
    packages=[
        "sansmic",
    ],  # Required
    ext_modules=ext_modules,
    cmdclass={"build_ext": build_ext},
)



# coding: utf-8
#
# Copyright (c) 2024 National Technology and Engineering Solutions of
# Sandia, LLC. Under the terms of Contract DE-NA0003525 with NTESS,
# the U.S. Government retains certain rights in this software.
#
# SPDX-License-Identifier: BSD-3-Clause.

"""Provide input and output functions for sansmic."""

import json
import logging
import os
import sys
import warnings
from dataclasses import fields
from enum import IntEnum

if sys.version_info[1] < 11:
    import tomli as tomllib
else:
    import tomllib

import numpy as np
import pandas as pd

try:
    import yaml
except ImportError as e:
    try:
        import ruamel.yaml as _yaml

        yaml = _yaml.YAML
    except ImportError:
        yaml = e

try:
    import h5py
except ImportError as e:
    h5py = e

from .model import (
    GeometryFormat,
    Results,
    Scenario,
    SimulationMode,
    StageDefinition,
    StopCondition,
    _OutDataBlock,
    _OutputData,
)

logger = logging.getLogger("sansmic")


def read_scenario(
    config_file: str, warn=True, strict=False, *, format=None
) -> Scenario:
    """Load a sansmic scenario file.

    Parameters
    ----------
    config_file : str
        filename or full path to a new-style configuration file
    warn : bool, optional
        emit warnings if there are extra, unrecognized options in the file, by default True
    strict : bool, optional
        raise errors if there are extra, unrecognized options in the file, by default False

    Returns
    -------
    Scenario
        the sansmic scenario

    Raises
    ------
    RuntimeError
        if there is an unrecognized option and ``strict`` is True, or if there are no
        stages defined.
    """
    logger.debug(
        "Attempting to read scenario {} in {} mode".format(
            config_file, "strict" if strict else "permissive"
        )
    )
    try:
        if config_file.lower().endswith(".json") or format == "json":
            with open(config_file, "r") as fin:
                data = json.load(fin)
        elif config_file.lower().endswith(".yaml") or format == "yaml":
            with open(config_file, "r") as fin:
                data = yaml.safe_load(fin)
        elif config_file.lower().endswith(".toml") or format == "toml":
            with open(config_file, "rb") as fin:
                data = tomllib.load(fin)
        elif config_file.lower().endswith(".dat") or format == "dat":
            with open(config_file, "r") as fin:
                data = read_dat(fin)
            return data
        else:
            with open(config_file, "rb") as fin:
                data = tomllib.load(fin)
    except:
        logger.error("Error reading scenario file {}".format(config_file))
        raise

    if "stages" not in data or len(data["stages"]) < 1:
        logger.error("No stages provided. Failed to load valid scenario.")
        raise RuntimeError("No stages provided.")
    return Scenario.from_dict(data)


def read_dat(str_or_buffer, *, ignore_errors=False) -> Scenario:
    """Read an old-style SANSMIC input DAT file.

    Parameters
    ----------
    str_or_buffer : str or stream
        The file or stream buffer to process.

    Returns
    -------
    Scenario
        the scenario derived from the DAT file

    """
    scenario = Scenario()
    with open(str_or_buffer, "r") if isinstance(
        str_or_buffer, str
    ) else str_or_buffer as fin:
        prefix = fin.name
        scenario.title = "Converted from {}".format(prefix)
        first = True
        logger.debug("Reading old DAT format file: {}".format(prefix))
        while True:
            stage = StageDefinition()
            # Stage - block 1
            title = fin.readline().strip()
            if title == "END":
                break
            logger.debug(" Record 1: Stage title")
            logger.debug("  title         = {}".format(title))
            # Stage - block 2
            (
                num_cells,
                mode,
                print_interval,
                subsequent,
                iResetGeo,
                rest_duration,
                num_coallescing,
                geometry_format,
                stop_value,
            ) = (
                fin.readline().strip().split()
            )
            logger.debug(" Record 2: General")
            logger.debug("  ndiv          = {}".format(num_cells))
            logger.debug("  leachtype     = {}".format(mode))
            logger.debug("  iprint        = {}".format(print_interval))
            logger.debug("  repeat        = {}".format(subsequent))
            logger.debug("  resetgeo[depr]= {}".format(iResetGeo))
            logger.debug("  iWait         = {}".format(rest_duration))
            logger.debug("  nco           = {}".format(num_coallescing))
            logger.debug("  idata         = {}".format(geometry_format))
            logger.debug("  ivol          = {}".format(stop_value))
            # Stage - block 3
            (
                cavern_height,
                injection_height,
                production_height,
                interface_height,
                ullage_standoff,
            ) = (
                fin.readline().strip().split()
            )
            logger.debug(" Record 3: Heights")
            logger.debug("  zmax          = {}".format(cavern_height))
            logger.debug("  zi            = {}".format(injection_height))
            logger.debug("  zp            = {}".format(production_height))
            logger.debug("  zb            = {}".format(interface_height))
            logger.debug("  zu            = {}".format(ullage_standoff))
            # Stage - block 4
            injection_rate = fin.readline().strip()
            logger.debug(" Record 4: Injection flow rates")
            logger.debug("  QI            = {}".format(injection_rate))
            # TODO: FIXME: handle injection tables? Or do we ignore for old
            # style inputs?
            #
            # Stage - block 5
            (
                inn_tbg_inside_radius,
                inn_tbg_outside_radius,
                out_csg_inside_radius,
                out_csg_outside_radius,
            ) = (
                fin.readline().strip().split()
            )
            logger.debug(" Record 5: Casing and tubing")
            logger.debug("  rpi           = {}".format(inn_tbg_inside_radius))
            logger.debug("  rpo           = {}".format(inn_tbg_outside_radius))
            logger.debug("  rcasi         = {}".format(out_csg_inside_radius))
            logger.debug("  rcaso         = {}".format(out_csg_outside_radius))
            # Stage - block 6
            injection_fluid_sg, cavern_sg = fin.readline().strip().split()
            logger.debug(" Record 6: Water and brine")
            logger.debug("  sginj         = {}".format(injection_fluid_sg))
            logger.debug("  sgcav         = {}".format(cavern_sg))
            # Stage - block 7
            timestep, injection_duration = fin.readline().strip().split()
            logger.debug(" Record 7: Timing")
            logger.debug("  dt            = {}".format(timestep))
            logger.debug("  tend          = {}".format(injection_duration))
            # Stage - block 8
            fill_rate, tDelay, coallescing_well_separation = (
                fin.readline().strip().split()
            )
            logger.debug(" Record 8: Oil fill rates")
            logger.debug("  qfil          = {}".format(fill_rate))
            logger.debug("  tdlay  [depr.]= {}".format(tDelay))
            logger.debug("  sep           = {}".format(coallescing_well_separation))
            if float(tDelay) != 0:
                logger.warning("The TDLAY option should not be used. Ignoring.")
            # First stage - block 9
            if first:
                logger.debug("First stage initialization")
                logger.debug(" Record 9: Geometry")
                geometry_data = list()
                first = False
                geometry_format = GeometryFormat(int(geometry_format))
                if geometry_format == GeometryFormat.RADIUS_LIST:
                    radii = list()
                    for _ in range(int(num_cells) + 1):
                        r = float(fin.readline().strip())
                        radii.append(r)
                    geometry_data = dict(radii=radii)
                elif geometry_format == GeometryFormat.VOLUME_LIST:
                    ndata = int(fin.readline().strip())
                    depths = list()
                    volumes = list()
                    for _ in range(ndata):
                        z = float(fin.readline().strip())
                        depths.append(z)
                    for _ in range(ndata):
                        v = float(fin.readline().strip())
                        volumes.append(v)
                    geometry_data = dict(depths=depths, volumes=volumes)
                elif geometry_format == GeometryFormat.VOLUME_TABLE:
                    ndata = int(fin.readline().strip())
                    depths = list()
                    volumes = list()
                    for _ in range(ndata):
                        z, v = fin.readline().strip().split()
                        depths.append(float(z))
                        volumes.append(float(v))
                    geometry_data = dict(depths=depths, volumes=volumes)
                elif geometry_format == GeometryFormat.VOLUME_TABLE:
                    ndata = int(fin.readline().strip())
                    depths = list()
                    radii = list()
                    for _ in range(ndata):
                        z, v = fin.readline().strip().split()
                        depths.append(float(z))
                        radii.append(float(v))
                    geometry_data = dict(depths=depths, radii=radii)

                dissolution_factor, insoluble_fraction, refdep, depth = (
                    fin.readline().strip().split()
                )
                logger.debug(" Record 10: Miscellaneous")
                logger.debug("  zdis          = {}".format(dissolution_factor))
                logger.debug("  zfin          = {}".format(insoluble_fraction))
                logger.debug("  refdep [depr.]= {}".format(refdep))
                logger.debug("  depth         = {}".format(depth))
                if refdep != depth:
                    logger.warning(
                        "The REFDEP is no longer used, only DEPTH. Ignoring REFDEP."
                    )
                if float(dissolution_factor) != 1.0:
                    logger.warning(
                        "The ZDIS should always be 1.0. This is a dangerous choice."
                    )
                scenario.geometry_format = geometry_format
                scenario.geometry_data = geometry_data
                scenario.num_cells = int(num_cells)
                scenario.coallescing_wells = int(num_coallescing)
                scenario.well_separation = float(coallescing_well_separation)
                df = float(dissolution_factor)
                if df != 1.0:
                    scenario.advanced.dissolution_factor = df
                scenario.insolubles_ratio = float(insoluble_fraction)
                scenario.floor_depth = float(depth)
                scenario.cavern_height = float(cavern_height)
                scenario.ullage_standoff = float(ullage_standoff)
            else:
                logger.debug("Subsequent stage initialization")
            if not first and scenario.num_cells != int(num_cells) and not ignore_errors:
                logger.critical("Old-style DAT has changes in NDIV throughout the file")
                raise TypeError("Invalid data in DAT file: NDIV not constant.")
            elif not first and int(iResetGeo) != 0 and not ignore_errors:
                logger.critical(
                    "The RESETGEO flag (item 5 on line 2 of stage {}) is set to 1. This is not valid in this version of SANSMIC. Please correct this and rerun.".format(
                        len(scenario.stages) + 1
                    )
                )
                raise TypeError("Invalid data in DAT file: RESETGEO not 0")
            if not first and not bool(1 - int(subsequent)):
                logger.warning(
                    "The REPEAT option is supposed to turn off the cavern SG; initial cavern SG set to None."
                )
                cavern_sg = 0.0
            stage.title = title
            stage.simulation_mode = SimulationMode(int(mode))
            stage.save_frequency = int(print_interval)
            stage.set_initial_conditions = bool(1 - int(subsequent))
            stage.rest_duration = float(rest_duration)
            stage.stop_value = float(stop_value)
            if stage.stop_value < 0:
                stage.stop_condition = StopCondition.DEPTH
            if stage.stop_value > 0:
                stage.stop_condition = StopCondition.VOLUME
            stage.brine_injection_depth = float(injection_height)
            stage.brine_production_depth = float(production_height)
            stage.brine_interface_depth = float(interface_height)
            stage.brine_injection_rate = float(injection_rate)
            stage.inner_tbg_inside_diam = float(inn_tbg_inside_radius) * 2
            stage.inner_tbg_outside_diam = float(inn_tbg_outside_radius) * 2
            stage.outer_csg_inside_diam = float(out_csg_inside_radius) * 2
            stage.outer_csg_outside_diam = float(out_csg_outside_radius) * 2
            stage.brine_injection_sg = float(injection_fluid_sg)
            stage.set_cavern_sg = float(cavern_sg)
            stage.solver_timestep = float(timestep)
            stage.injection_duration = float(injection_duration)
            stage.product_injection_rate = float(fill_rate)
            scenario.stages.append(stage)
            logger.debug("Finished reading stage")
    return scenario


def write_scenario(scenario: Scenario, filename: str, *, redundant=False, format=None):
    """Write a new-style SANSMIC scenario file (preferred extension is .toml)"""
    sdict = scenario.to_dict(redundant)
    keys = [k for k in sdict.keys()]
    for k in keys:
        if not redundant and k == "coallescing-wells" and sdict[k] == 1:
            sdict[k] = None
            sdict["well-separation"] = None
        if sdict[k] is None:
            if redundant:
                sdict[k] = ""
            else:
                del sdict[k]
        elif isinstance(sdict[k], IntEnum):
            sdict[k] = sdict[k].name.lower().replace("_", "-")
    for ct, s in enumerate(sdict["stages"]):
        if (
            not redundant
            and scenario.stages[ct].stop_condition == StopCondition.DURATION
        ):
            del s["stop-condition"]
            del s["stop-value"]
        if scenario.stages[ct].set_initial_conditions is False and ct > 0:
            s["set-cavern-sg"] = None
            s["brine-interface-depth"] = None
            s["set-initial-conditions"] = None
        elif scenario.stages[ct].set_initial_conditions is None and ct > 0:
            s["set-cavern-sg"] = None
            if not s.get("brine-interface-depth", None) and redundant:
                s["brine-interface-depth"] = 0
        keys = [k for k in s.keys()]
        for k in keys:
            if (
                not redundant
                and k in scenario.defaults
                and scenario.defaults.get(k, None) == s[k]
            ):
                del s[k]
                continue
            if s[k] is None:
                if redundant:
                    s[k] = ""
                else:
                    del s[k]
            elif isinstance(s[k], IntEnum):
                s[k] = s[k].name.lower().replace("_", "-")
        if not redundant and ct == 0 and "set-initial-conditions" in s:
            del s["set-initial-conditions"]
    name, ext = os.path.splitext(filename)
    with open(filename, "w") as fout:
        if ext.lower() == ".toml" or format == "toml":
            for k, v in sdict.items():
                if k in ["stages", "defaults", "advanced"]:
                    continue
                elif k == "geometry-data" and isinstance(v, dict):
                    for k2, v2 in v.items():
                        fout.write("geometry-data.{} = {}\n".format(k2, v2))
                    continue
                if isinstance(v, bool):
                    v = str(v).lower()
                elif isinstance(v, str):
                    v = repr(v)
                elif isinstance(v, IntEnum):
                    v = repr(v.name.lower().replace("_", "-"))
                fout.write("{} = {}\n".format(k, v))

            if "advanced" in sdict:
                fout.write("\n[advanced]\n")
                for k, v in sdict["advanced"].items():
                    if isinstance(v, bool):
                        v = str(v).lower()
                    elif isinstance(v, str):
                        v = repr(v)
                    elif isinstance(v, IntEnum):
                        v = repr(v.name.lower().replace("_", "-"))
                    fout.write("{} = {}\n".format(k, v))

            if len(sdict["defaults"]) > 0:
                fout.write("\n[defaults]\n")
                for k, v in sdict["defaults"].items():
                    if isinstance(v, bool):
                        v = str(v).lower()
                    elif isinstance(v, str):
                        v = repr(v)
                    elif isinstance(v, IntEnum):
                        v = repr(v.name.lower().replace("_", "-"))
                    fout.write("{} = {}\n".format(k, v))

            for s in sdict["stages"]:
                fout.write("\n[[stages]]\n")
                for k, v in s.items():
                    if isinstance(v, bool):
                        v = str(v).lower()
                    elif isinstance(v, str):
                        v = repr(v)
                    elif isinstance(v, IntEnum):
                        v = repr(v.name.lower().replace("_", "-"))
                    fout.write("{} = {}\n".format(k, v))
        elif ext.lower() == ".json" or format == "json":
            json.dump(sdict, fout)
        elif ext.lower() == ".yaml" or format == "yaml":
            yaml.dump(sdict, fout)
        else:
            logger.critical(
                'Unknown file extension ({}) and invalid format ({}). Valid extensions/formats are "toml", "yaml" and "json".'.format(
                    ext, format
                )
            )
            raise RuntimeError("Unknown file format for scenario output")


def read_classic_out_ddl(file_prefix):
    """
    Read in the data from ".out" and ".ddl" files from a classic SANSMIC run.

    Parameters
    ----------
    file_prefix : [type]
        [description]

    Returns
    -------
    [type]
        [description]
    """

    started = False
    dataline = False
    skip1 = False
    steps = list()
    step = None
    phaseline = False

    r_0 = list()
    z_0 = list()
    h_0 = list()

    with open(file_prefix + ".ddl", "r") as fddl:
        for ct, line in enumerate(fddl.readlines()):
            if ct == 0:
                continue
            words = line.split()
            idx = int(words[0])
            z = float(words[1])
            r = float(words[2])
            r_0.insert(0, r)
            z_0.insert(0, z)
            if idx == 1:
                break
    for z in z_0:
        h = z - z_0[0]
        h_0.append(h)

    with open(file_prefix + ".out", "r") as fout:
        for line in fout.readlines():
            words = line.split()
            if skip1:
                skip1 = False
                continue
            if not skip1 and len(words) == 0:
                continue
            if not started and words[0] == "TIME=":
                started = True
                step = _OutDataBlock()
            elif not started:
                continue
            if words[0] == "TIME=":
                step.t = float(words[1]) * 24.0
                step.dt = float(words[7]) * 24.0
                step.t_prime = float(words[11]) * 24.0
                phaseline = True
            elif phaseline and words[0] == "INJECTION":
                step.phase = 1
                step.stage = int(words[2])
                phaseline = False
            elif phaseline and words[0] == "WORKOVER":
                step.phase = 0
                step.stage = int(words[2])
                phaseline = False
            elif words[0] == "I(INJ),":
                step.injCell = int(words[4])
                step.prodCell = int(words[5])
                step.obiCell = int(words[6])
                step.plumCell = int(words[7])
            elif words[0] == "H(INJ),":
                step.z_inj = z_0[0] - float(words[4])
                step.z_prod = z_0[0] - float(words[5])
                # step.z_obi = z_0[0] - float(words[6])
                step.z_plm = z_0[0] - float(words[7])
            elif words[0] == "Ljet(ft),":
                step.l_jet = float(words[4])
                step.r_jet = float(words[5])
                step.u_jet = float(words[6])
                skip1 = True
            elif words[0] == "HEIGHT":
                skip1 = True
                dataline = True
            elif dataline:
                cellid = int(line[0:5])
                if cellid == 1:
                    dataline = False
                    skip1 = True
                step.idx.insert(0, cellid)
                step.h.insert(0, float(line[5:16]))
                step.r.insert(0, float(line[16:27]))
                step.dr_0.insert(0, float(line[27:38]))
                step.sg.insert(0, float(line[38:49]))
                step.theta.insert(0, float(line[49:60]))
                step.Q_inj.insert(0, float(line[60:71]))
                step.V.insert(0, float(line[71:82]))
                step.f_dis.insert(0, float(line[82:93]))
                step.f_flag.insert(0, int(line[93:98]))
                step.xincl.insert(0, float(line[98:109]))
                step.amd.insert(0, float(line[109:120]))
                step.D_coeff.insert(0, float(line[120:131]))
                step.dC_dz.insert(0, float(line[131:142]))
                step.C_old.insert(0, float(line[142:153]))
                step.C_new.insert(0, float(line[153:164]))
                step.dC.insert(0, float(line[164:175]))
                step.dr.insert(0, float(line[175:186]))
                step.C_plm.insert(0, float(line[186:197]))
                step.u_plm.insert(0, float(line[197:208]))
                step.r_plm.insert(0, float(line[208:]))
            elif words[0] == "TOTAL":
                step.V_cavTot = float(line[25:36])
            elif words[0] == "BRINE" and words[1] == "OUT":
                step.Q_out = float(line[25:36])
            elif words[0] == "OUTLET":
                step.sg_out = float(line[25:36])
            elif words[0] == "VOLUME":
                step.V_insolTot = float(line[25:36])
            elif words[0] == "INSOL":
                step.z_insol = z_0[0] - float(line[25:36])
            elif words[0] == "BLANKET":
                step.z_obi = z_0[0] - float(line[25:36])
            elif words[0] == "VOL":
                step.V_insolVent = float(line[25:36])
            elif words[0] == "BRINE" and words[1] == "VOLUME":
                step.V_brine = float(line[25:36])
            elif words[0] == "ULLAGE":
                step.V_ullage = float(line[25:36])
            elif words[0] == "USEABLE":
                step.V_usable = float(line[25:36])
            elif words[0] == "CFAC":
                step.err = float(line[25:36])
            elif line.startswith("----------"):
                # TODO: add datablock to list
                steps.append(step)
                started = False
            else:
                print("unrecognized line: ", line)
    combo = _OutputData()
    combo.z_0 = z_0
    combo.h_0 = h_0
    combo.r_0 = r_0
    combo.V_injTot = list()
    combo.V_fillTot = list()
    combo.sg_cavAve = list()

    for step in steps:
        combo.V_injTot.append(np.nan)
        combo.V_fillTot.append(np.nan)
        combo.sg_cavAve.append(np.nan)
        for key in step.__dict__.keys():
            if not hasattr(combo, key):
                continue
            else:
                tmp = getattr(combo, key)
                if tmp is None:
                    setattr(combo, key, list())
                if key == "t":
                    val = step.t + step.t_prime
                else:
                    val = getattr(step, key)
                getattr(combo, key).append(val)

    res = Results(combo)
    return res


def read_tst_file(file_prefix: str):
    """Read a .tst output file into a DataFrame.

    Parameters
    ----------
    file_prefix : str
        Filename without extension.
    """
    Q_fill = list()
    t = list()
    V_cavTot = list()
    err = list()
    sg_out = list()
    sg_cavAve = list()
    V_insolTot = list()
    z_insol = list()
    z_obi = list()
    V_insolVent = list()
    V_ullage = list()
    V_usable = list()
    Q_inj = list()
    Q_fill = list()
    V_injTot = list()
    V_fillTot = list()
    with open(file_prefix + ".tst", "r") as fout:
        for line in fout.readlines():
            words = line.split()
            if len(words) == 0:
                continue
            if words[0] in ["#", "t", "(d)", "File=", "TIME", "Days", "---", "==="]:
                continue
            t.append(float(words[0]))
            V_cavTot.append(float(words[1]))
            err.append(float(words[2]))
            sg_out.append(float(words[3]))
            sg_cavAve.append(float(words[4]))
            V_insolTot.append(float(words[5]))
            z_insol.append(float(words[6]))
            z_obi.append(float(words[7]))
            V_insolVent.append(float(words[8]))
            V_ullage.append(float(words[9]))
            V_usable.append(float(words[10]))
            Q_inj.append(float(words[11]))
            V_injTot.append(float(words[12]))
            Q_fill.append(float(words[13]))
            V_fillTot.append(float(words[14]))
    return pd.DataFrame.from_dict(
        dict(
            t_d=t,
            V_cav=V_cavTot,
            err_ode=err,
            sg_out=sg_out,
            sg_ave=sg_cavAve,
            V_insol=V_insolTot,
            z_insol=z_insol,
            z_obi=z_obi,
            V_vented=V_insolVent,
            Q_inj=Q_inj,
            Q_fill=Q_fill,
            V_inj=V_injTot,
            V_fill=V_fillTot,
        )
    )


def write_hdf_results(results: Results, filename: str, **kwargs):
    """Write results to an HDF5 file.

    Parameters
    ----------
    results : Results
        The results object to write
    filename : str
        The filename to write the results to.

    Keyword Arguments
    -----------------
    kwargs : additional keyword arguments
        See :meth:`~sansmic.model.Results.to_hdf` for keyword arguments.

    """
    results.to_hdf(filename)


def read_hdf_results(filename: str) -> Results:
    """Read results from an HDF5 file.

    Parameters
    ----------
    filename : str
        Filename to read the results from.

    """
    results = Results.from_hdf(filename)
    return results


def write_json_results(results: Results, filename: str, **kwargs):
    """Write results to a JSON file.

    Parameters
    ----------
    results : Results
        The results object to write
    filename : str
        The filename to write the results to.

    Keyword Arguments
    -----------------
    kwargs : additional keyword arguments
        See :meth:`json.dump` for valid keyword arguments

    """
    with open(filename, "w") as f:
        json.dump(results.to_dict(), f, **kwargs)


def read_json_results(filename: str):
    """Read results from a JSON file.

    Parameters
    ----------
    filename : str
        Filename to read the results from.
    """
    with open(filename, "r") as f:
        d = json.load(f)
    return Results.from_dict(d)


def write_csv_results(results: Results, prefix: str):
    """Write results to CSV files.

    Parameters
    ----------
    results : Results
        The results to write out.
    prefix : str
        Write results files in CSV formats.
    """
    with open(prefix + "-summary.csv", "w") as f:
        results.df_t_1D.to_csv(f, lineterminator="\n", index=False)
    with open(prefix + "-radius.csv", "w") as f:
        df = results.radius
        df.index = results.depths
        df = df.T
        df.index = results.time
        df = df.T
        df.to_csv(f, lineterminator="\n")
    with open(prefix + "-density.csv", "w") as f:
        df = results.cell_sg
        df.index = results.depths
        df = df.T
        df.index = results.time
        df = df.T
        df.to_csv(f, lineterminator="\n")
    with open(prefix + "-wall-angle.csv", "w") as f:
        df = results.wall_angle
        df.index = results.depths
        df = df.T
        df.index = results.time
        df = df.T
        df.to_csv(f, lineterminator="\n")
    with open(prefix + "-dr_dt.csv", "w") as f:
        df = results.rate_of_change_in_radius
        df.index = results.depths
        df = df.T
        df.index = results.time
        df = df.T
        df.to_csv(f, lineterminator="\n")
    with open(prefix + "-dC_dt.csv", "w") as f:
        df = results.rate_of_change_in_sg
        df.index = results.depths
        df = df.T
        df.index = results.time
        df = df.T
        df.to_csv(f, lineterminator="\n")
    with open(prefix + "-dC_dz.csv", "w") as f:
        df = results.vertical_diffusion_rate
        df.index = results.depths
        df = df.T
        df.index = results.time
        df = df.T
        df.to_csv(f, lineterminator="\n")



# coding: utf-8
#
# Copyright (c) 2024 National Technology and Engineering Solutions of
# Sandia, LLC. Under the terms of Contract DE-NA0003525 with NTESS,
# the U.S. Government retains certain rights in this software.
#
# SPDX-License-Identifier: BSD-3-Clause.

"""SANSMIC package"""
__version__ = "1.0.0-rc.1"
__copyright__ = f"""sansmic ver. {__version__}
Copyright (c) 2024 National Technology and Engineering Solutions of Sandia,
LLC (NTESS). Under the terms of Contract DE-NA0003525 with NTESS, the U.S.
Government retains certain rights in this software."""
__license__ = f"""Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this
   list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice,
   this list of conditions and the following disclaimer in the documentation
   and/or other materials provided with the distribution.

3. Neither the name of the copyright holder nor the names of its
   contributors may be used to endorse or promote products derived from
   this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."""


from . import app, empirical, io, model
